########################################################################
# File name: roster.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp.roster

from framework import Example, exec_example


class Roster(Example):
    def _print_item(self, item):
        print("  entry {}:".format(item.jid))
        print("    name={!r}".format(item.name))
        print("    subscription={!r}".format(item.subscription))
        print("    ask={!r}".format(item.ask))
        print("    approved={!r}".format(item.approved))

    def _on_initial_roster(self):
        for group, items in self.roster.groups.items():
            print("group {}:".format(group))
            for item in items:
                self._print_item(item)

        print("ungrouped items:")
        for item in self.roster.items.values():
            if not item.groups:
                self._print_item(item)

        self.done_event.set()

    def make_simple_client(self):
        client = super().make_simple_client()
        self.roster = client.summon(aioxmpp.RosterClient)
        self.roster.on_initial_roster_received.connect(
            self._on_initial_roster,
        )
        self.done_event = asyncio.Event()

        return client

    @asyncio.coroutine
    def run_simple_example(self):
        done, pending = yield from asyncio.wait(
            [
                self.sigint_event.wait(),
                self.done_event.wait()
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )
        for fut in pending:
            fut.cancel()

    @asyncio.coroutine
    def run_example(self):
        self.sigint_event = self.make_sigint_event()
        yield from super().run_example()


if __name__ == "__main__":
    exec_example(Roster())



########################################################################
# File name: test_types.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import contextlib
import decimal
import fractions
import inspect
import ipaddress
import itertools
import unittest
import unittest.mock
import warnings

from enum import Enum, IntEnum

import pytz

from datetime import datetime, date, time

import aioxmpp.xso as xso
import aioxmpp.structs as structs


class Unknown(unittest.TestCase):
    def test_init(self):
        u = xso.Unknown(unittest.mock.sentinel.value)
        self.assertEqual(
            u.value,
            unittest.mock.sentinel.value,
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            xso.Unknown()

    def test_value_not_settable(self):
        u = xso.Unknown(unittest.mock.sentinel.value)
        with self.assertRaises(AttributeError):
            u.value = "foobar"

    def test_hash_equal_to_value_hash(self):
        values = [
            None,
            "foobar",
            object(),
            10,
            10.2,
        ]

        for value in values:
            u = xso.Unknown(value)
            self.assertEqual(hash(u), hash(value))

    def test_equality(self):
        values = [
            None,
            "foobar",
            object(),
            10,
            10.2,
        ]

        for v1, v2 in itertools.product(values, values):
            if v1 == v2:
                self.assertTrue(xso.Unknown(v1) == xso.Unknown(v2))
                self.assertFalse(xso.Unknown(v1) != xso.Unknown(v2))
            else:
                self.assertFalse(xso.Unknown(v1) == xso.Unknown(v2))
                self.assertTrue(xso.Unknown(v1) != xso.Unknown(v2))
                self.assertFalse(v1 == xso.Unknown(v2))
                self.assertFalse(xso.Unknown(v2) == v1)
                self.assertTrue(v1 != xso.Unknown(v2))
                self.assertTrue(xso.Unknown(v2) != v1)

    def test_repr(self):
        values = [
            None,
            "foobar",
            object(),
            10,
            10.2,
        ]

        for v in values:
            self.assertEqual(
                repr(xso.Unknown(v)),
                "<Unknown: {!r}>".format(v),
            )


class TestAbstractCDataType(unittest.TestCase):
    class DummyType(xso.AbstractCDataType):
        def parse(self, v):
            pass

    def test_is_abstract(self):
        self.assertIsInstance(
            xso.AbstractCDataType,
            abc.ABCMeta)
        with self.assertRaises(TypeError):
            xso.AbstractCDataType()

    def test_parse_method(self):
        self.assertTrue(inspect.isfunction(xso.AbstractCDataType.parse))

    def test_format_method(self):
        self.assertTrue(inspect.isfunction(xso.AbstractCDataType.format))
        self.assertEqual(
            "foo",
            self.DummyType().format("foo"))
        self.assertEqual(
            "23",
            self.DummyType().format(23))


class TestAbstractElementType(unittest.TestCase):
    class DummyType(xso.AbstractElementType):
        def unpack(self, obj):
            pass

        def pack(self, v):
            pass

        def get_xso_types(self):
            pass

    def test_is_abstract(self):
        self.assertIsInstance(
            xso.AbstractElementType,
            abc.ABCMeta)
        with self.assertRaises(TypeError):
            xso.AbstractElementType()

    def test_unpack_method(self):
        self.assertTrue(inspect.isfunction(xso.AbstractElementType.unpack))

    def test_coerce_method(self):
        self.assertTrue(inspect.isfunction(xso.AbstractElementType.coerce))
        self.assertEqual(
            self.DummyType().coerce(unittest.mock.sentinel.value),
            unittest.mock.sentinel.value
        )

    def test_get_xso_types(self):
        self.assertTrue(inspect.isfunction(
            xso.AbstractElementType.get_xso_types
        ))

    def test_pack_method(self):
        self.assertTrue(inspect.isfunction(xso.AbstractElementType.pack))


class TestStringType(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.String(),
            xso.AbstractCDataType)

    def test_parse(self):
        t = xso.String()
        self.assertEqual(
            "foo",
            t.parse("foo"))

    def test_format(self):
        t = xso.String()
        self.assertEqual(
            "foo",
            t.format("foo"))

    def test_coerce_passes_string(self):
        t = xso.String()
        s = "foobar"
        self.assertIs(s, t.coerce(s))

    def test_coerce_rejects_non_strings(self):
        t = xso.String()

        values = [
            1.2,
            decimal.Decimal("1"),
            fractions.Fraction(1, 1),
            [],
            (),
            1.
        ]

        for value in values:
            with self.assertRaisesRegex(TypeError, "must be a str"):
                t.coerce(value)

    def test_coerce_stringprep(self):
        prepfunc = unittest.mock.Mock()
        t = xso.String(prepfunc=prepfunc)

        result = t.coerce("foobar")

        self.assertSequenceEqual(
            [
                unittest.mock.call("foobar"),
            ],
            prepfunc.mock_calls
        )

        self.assertEqual(
            prepfunc(),
            result,
        )

    def test_parse_stringprep(self):
        prepfunc = unittest.mock.Mock()
        t = xso.String(prepfunc=prepfunc)

        result = t.parse("foobar")

        self.assertSequenceEqual(
            [
                unittest.mock.call("foobar"),
            ],
            prepfunc.mock_calls
        )

        self.assertEqual(
            prepfunc(),
            result,
        )


class TestIntegerType(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.Integer(),
            xso.AbstractCDataType)

    def test_parse(self):
        t = xso.Integer()
        self.assertEqual(
            123,
            t.parse("123"))

    def test_parse_failure(self):
        t = xso.Integer()
        with self.assertRaises(ValueError):
            t.parse("123f")

    def test_format(self):
        t = xso.Integer()
        self.assertEqual(
            "123",
            t.format(123))

    def test_coerce_passes_integral_numbers(self):
        t = xso.Integer()

        values = [-2, 0, 1, 2, 3, 4, 100]

        for value in values:
            self.assertIs(value, t.coerce(value))

        import random
        value = random.randint(1, 1e10)
        self.assertIs(value, t.coerce(value))
        value = -value
        self.assertIs(value, t.coerce(value))

    def test_coerce_requires_integral_number(self):
        t = xso.Integer()

        values = [
            1.2,
            "1",
            decimal.Decimal("1"),
            fractions.Fraction(1, 1),
            "foo",
            [],
            (),
            1.
        ]

        for value in values:
            with self.assertRaisesRegex(
                    TypeError,
                    "must be integral number"):
                t.coerce(value)


class TestFloatType(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.Float(),
            xso.AbstractCDataType)

    def test_parse(self):
        t = xso.Float()
        self.assertEqual(
            123.3,
            t.parse("123.3"))

    def test_parse_failure(self):
        t = xso.Float()
        with self.assertRaises(ValueError):
            t.parse("123.3f")

    def test_format(self):
        t = xso.Float()
        self.assertEqual(
            "123.3",
            t.format(123.3))

    def test_coerce_passes_real_numbers(self):
        t = xso.Float()

        values = [
            # decimal.Decimal("1.23"),
            fractions.Fraction(1, 9),
            1.234,
            20,
            -1,
            -3.4,
        ]

        for value in values:
            self.assertEqual(
                float(value),
                t.coerce(value)
            )

    def test_coerce_passes_decimal(self):
        t = xso.Float()

        values = [
            decimal.Decimal("1.23"),
        ]

        for value in values:
            self.assertEqual(
                float(value),
                t.coerce(value)
            )

    def test_coerce_requires_float_number(self):
        t = xso.Float()

        values = [
            "foo",
            [],
            ()
        ]

        for value in values:
            with self.assertRaisesRegex(
                    TypeError,
                    "must be real number"):
                t.coerce(value)


class TestBoolType(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.Bool(),
            xso.AbstractCDataType)

    def test_parse(self):
        t = xso.Bool()
        self.assertTrue(t.parse("true"))
        self.assertTrue(t.parse("1"))
        self.assertTrue(t.parse("  true  "))
        self.assertTrue(t.parse(" 1 "))
        self.assertFalse(t.parse("false"))
        self.assertFalse(t.parse("0"))
        self.assertFalse(t.parse("  false "))
        self.assertFalse(t.parse(" 0 "))

    def test_parse_failure(self):
        t = xso.Bool()
        with self.assertRaises(ValueError):
            t.parse("foobar")
        with self.assertRaises(ValueError):
            t.parse("truefoo")
        with self.assertRaises(ValueError):
            t.parse("0foo")

    def test_format(self):
        t = xso.Bool()
        self.assertEqual(
            "true",
            t.format(True))
        self.assertEqual(
            "false",
            t.format(False))

    def test_coerce_anything(self):
        t = xso.Bool()
        mock = unittest.mock.MagicMock()

        result = mock.__bool__()
        mock.reset_mock()

        self.assertEqual(
            result,
            t.coerce(mock))

        mock.__bool__.assert_called_once_with()


class TestDateTimeType(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.DateTime(),
            xso.AbstractCDataType)

    def test_parse_example(self):
        t = xso.DateTime()
        self.assertEqual(
            datetime(2014, 1, 26, 19, 40, 10, tzinfo=pytz.utc),
            t.parse("2014-01-26T19:40:10Z"))

    def test_parse_timezoned(self):
        t = xso.DateTime()
        self.assertEqual(
            datetime(2014, 1, 26, 19, 40, 10, tzinfo=pytz.utc),
            t.parse("2014-01-26T20:40:10+01:00"))

    def test_parse_local(self):
        t = xso.DateTime()
        self.assertEqual(
            datetime(2014, 1, 26, 20, 40, 10),
            t.parse("2014-01-26T20:40:10"))

    def test_parse_milliseconds(self):
        t = xso.DateTime()
        self.assertEqual(
            datetime(2014, 1, 26, 20, 40, 10, 123456),
            t.parse("2014-01-26T20:40:10.123456"))

    def test_parse_milliseconds_timezoned(self):
        t = xso.DateTime()
        self.assertEqual(
            datetime(2014, 1, 26, 19, 40, 10, 123456, tzinfo=pytz.utc),
            t.parse("2014-01-26T20:40:10.123456+01:00"))

    def test_parse_need_time(self):
        t = xso.DateTime()
        with self.assertRaises(ValueError):
            t.parse("2014-01-26")

    def test_parse_need_date(self):
        t = xso.DateTime()
        with self.assertRaises(ValueError):
            t.parse("20:40:10")

    def test_format_timezoned(self):
        t = xso.DateTime()
        self.assertEqual(
            "2014-01-26T19:40:10Z",
            t.format(datetime(2014, 1, 26, 19, 40, 10, tzinfo=pytz.utc))
        )

    def test_format_timezoned_microseconds(self):
        t = xso.DateTime()
        self.assertEqual(
            "2014-01-26T19:40:10.1234Z",
            t.format(datetime(2014, 1, 26, 19, 40, 10, 123400,
                              tzinfo=pytz.utc))
        )

    def test_format_naive(self):
        t = xso.DateTime()
        self.assertEqual(
            "2014-01-26T19:40:10",
            t.format(datetime(2014, 1, 26, 19, 40, 10))
        )

    def test_format_naive_microseconds(self):
        t = xso.DateTime()
        self.assertEqual(
            "2014-01-26T19:40:10.1234",
            t.format(datetime(2014, 1, 26, 19, 40, 10, 123400))
        )

    def test_format_timezoned_nonutc(self):
        t = xso.DateTime()
        self.assertEqual(
            "2014-01-26T19:40:10Z",
            t.format(pytz.timezone("Europe/Berlin").localize(
                datetime(2014, 1, 26, 20, 40, 10)
            ))
        )

    def test_parse_xep0082_examples(self):
        t = xso.DateTime()
        self.assertEqual(
            t.parse("1969-07-21T02:56:15Z"),
            datetime(1969, 7, 21, 2, 56, 15, tzinfo=pytz.utc)
        )
        self.assertEqual(
            t.parse("1969-07-20T21:56:15-05:00"),
            datetime(1969, 7, 21, 2, 56, 15, tzinfo=pytz.utc)
        )

    def test_parse_legacy_format(self):
        t = xso.DateTime()
        self.assertEqual(
            t.parse("19690721T02:56:15"),
            datetime(1969, 7, 21, 2, 56, 15, tzinfo=pytz.utc)
        )

    def test_emit_legacy_format_with_switch(self):
        t = xso.DateTime(legacy=True)
        self.assertEqual(
            "19690721T02:56:15",
            t.format(datetime(1969, 7, 21, 2, 56, 15, tzinfo=pytz.utc))
        )
        self.assertEqual(
            "20140126T19:40:10",
            t.format(pytz.timezone("Europe/Berlin").localize(
                datetime(2014, 1, 26, 20, 40, 10)
            ))
        )

    def test_require_datetime(self):
        t = xso.DateTime()

        values = [
            1,
            "foo",
            "2014-01-26T19:47:10Z",
            12.3,
        ]

        for value in values:
            with self.assertRaisesRegex(
                    TypeError,
                    "must be a datetime object"):
                t.coerce(value)

    def test_pass_datetime(self):
        t = xso.DateTime()

        dt = datetime.utcnow()
        self.assertIs(
            dt,
            t.coerce(dt)
        )


class TestDate(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.Date(),
            xso.AbstractCDataType)

    def test_parse(self):
        t = xso.Date()
        self.assertEqual(
            t.parse("1776-07-04"),
            date(1776, 7, 4),
        )

    def test_format(self):
        t = xso.Date()
        self.assertEqual(
            t.format(date(1776, 7, 4)),
            "1776-07-04",
        )

    def test_coerce_rejects_datetime(self):
        t = xso.Date()
        with self.assertRaisesRegex(
                TypeError,
                "must be a date object"):
            t.coerce(datetime.utcnow())

    def test_coerce_rejects_time(self):
        t = xso.Date()
        with self.assertRaisesRegex(
                TypeError,
                "must be a date object"):
            t.coerce(datetime.utcnow().time())

    def test_coerce_accepts_date(self):
        t = xso.Date()
        v = datetime.utcnow().date()
        self.assertEqual(t.coerce(v), v)


class TestTime(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.Time(),
            xso.AbstractCDataType)

    def test_parse_example(self):
        t = xso.Time()
        self.assertEqual(
            time(19, 40, 10, tzinfo=pytz.utc),
            t.parse("19:40:10Z"))

    def test_parse_timezoned(self):
        t = xso.Time()
        self.assertEqual(
            time(19, 40, 10, tzinfo=pytz.utc),
            t.parse("20:40:10+01:00"))

    def test_parse_local(self):
        t = xso.Time()
        self.assertEqual(
            time(20, 40, 10),
            t.parse("20:40:10"))

    def test_parse_milliseconds(self):
        t = xso.Time()
        self.assertEqual(
            time(20, 40, 10, 123456),
            t.parse("20:40:10.123456"))

    def test_parse_milliseconds_timezoned(self):
        t = xso.Time()
        self.assertEqual(
            time(19, 40, 10, 123456, tzinfo=pytz.utc),
            t.parse("20:40:10.123456+01:00"))

    def test_format_timezoned(self):
        t = xso.Time()
        self.assertEqual(
            "19:40:10Z",
            t.format(time(19, 40, 10, tzinfo=pytz.utc))
        )

    def test_format_timezoned_microseconds(self):
        t = xso.Time()
        self.assertEqual(
            "19:40:10.1234Z",
            t.format(time(19, 40, 10, 123400,
                          tzinfo=pytz.utc))
        )

    def test_format_naive(self):
        t = xso.Time()
        self.assertEqual(
            "19:40:10",
            t.format(time(19, 40, 10))
        )

    def test_format_naive_microseconds(self):
        t = xso.Time()
        self.assertEqual(
            "19:40:10.1234",
            t.format(time(19, 40, 10, 123400))
        )

    def test_coerce_rejects_non_utc_timezone(self):
        t = xso.Time()
        with self.assertRaisesRegex(
                ValueError,
                "time must have UTC timezone or none at all"):
            t.coerce(pytz.timezone("Europe/Berlin").localize(
                datetime(2014, 1, 26, 20, 40, 10)
            ).timetz())

    def test_coerce_accepts_naive_timezone(self):
        t = xso.Time()
        v = time(20, 40, 10)
        result = t.coerce(v)
        self.assertEqual(v, result)

    def test_coerce_accepts_utc_timezone(self):
        t = xso.Time()
        v = time(20, 40, 10, tzinfo=pytz.utc)
        result = t.coerce(v)
        self.assertEqual(v, result)

    def test_coerce_rejects_datetime(self):
        t = xso.Time()
        with self.assertRaisesRegex(
                TypeError,
                "must be a time object"):
            t.coerce(datetime.utcnow())

    def test_coerce_rejects_date(self):
        t = xso.Time()
        with self.assertRaisesRegex(
                TypeError,
                "must be a time object"):
            t.coerce(datetime.utcnow().date())


class TestBase64Binary(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.Base64Binary(),
            xso.AbstractCDataType)

    def test_parse(self):
        t = xso.Base64Binary()
        self.assertEqual(
            b"fnord",
            t.parse("Zm5vcmQ=")
        )

    def test_parse_empty(self):
        t = xso.Base64Binary()
        self.assertEqual(
            b"",
            t.parse("")
        )
        self.assertEqual(
            b"",
            t.parse("=")
        )

    def test_format(self):
        t = xso.Base64Binary()
        self.assertEqual(
            "Zm5vcmQ=",
            t.format(b"fnord")
        )

    def test_format_empty_default(self):
        t = xso.Base64Binary()
        self.assertEqual(
            "",
            t.format(b"")
        )

    def test_format_empty_with_empty_as_equal_flag(self):
        t = xso.Base64Binary(empty_as_equal=True)
        self.assertEqual(
            "=",
            t.format(b"")
        )

    def test_format_long(self):
        t = xso.Base64Binary()
        self.assertEqual(
            "Zm5vcmRmbm9yZGZub3JkZm5vcmRmbm9yZGZub3JkZm5vcmRmbm9yZG"
            "Zub3JkZm5vcmRmbm9yZGZub3JkZm5vcmRmbm9yZGZub3JkZm5vcmRm"
            "bm9yZGZub3JkZm5vcmRmbm9yZA==",
            t.format(b"fnord"*20)
        )

    def test_coerce_rejects_int(self):
        t = xso.Base64Binary()
        with self.assertRaisesRegex(TypeError,
                                    "must be convertible to bytes"):
            t.coerce(12)

    def test_coerce_accepts_bytes_bytearray_array(self):
        t = xso.Base64Binary()

        import array
        array_value = array.array("h")
        array_value.append(1234)
        array_value.append(5678)
        array_value.append(910)

        values = [
            b"foobar",
            bytearray(b"baz"),
        ]

        for value in values:
            result = t.coerce(value)
            self.assertEqual(
                bytes(value),
                result
            )
            self.assertIsInstance(
                result,
                bytes
            )

    def test_coerce_passes_bytes(self):
        t = xso.Base64Binary()

        value = b"foo"

        self.assertIs(
            value,
            t.coerce(value)
        )


class TestHexBinary(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.HexBinary(),
            xso.AbstractCDataType)

    def test_parse(self):
        t = xso.HexBinary()
        self.assertEqual(
            b"fnord",
            t.parse("666e6f7264")
        )

    def test_format(self):
        t = xso.HexBinary()
        self.assertEqual(
            "666e6f7264",
            t.format(b"fnord")
        )

    def test_coerce_rejects_int(self):
        t = xso.HexBinary()
        with self.assertRaisesRegex(TypeError,
                                    "must be convertible to bytes"):
            t.coerce(12)

    def test_coerce_accepts_bytes_bytearray_array(self):
        t = xso.HexBinary()

        import array
        array_value = array.array("h")
        array_value.append(1234)
        array_value.append(5678)
        array_value.append(910)

        values = [
            b"foobar",
            bytearray(b"baz"),
        ]

        for value in values:
            result = t.coerce(value)
            self.assertEqual(
                bytes(value),
                result
            )
            self.assertIsInstance(
                result,
                bytes
            )

    def test_coerce_passes_bytes(self):
        t = xso.HexBinary()

        value = b"foo"

        self.assertIs(
            value,
            t.coerce(value)
        )


class TestJID(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.JID(),
            xso.AbstractCDataType)

    def test_parse(self):
        t = xso.JID()
        self.assertEqual(
            structs.JID("foo", "example.test", "bar"),
            t.parse("foo@example.test/bar")
        )

    def test_parse_uses_nonstrict_by_default(self):
        with unittest.mock.patch("aioxmpp.structs.JID") as JID:
            t = xso.JID()
            result = t.parse(unittest.mock.sentinel.jidstr)

        JID.fromstr.assert_called_with(
            unittest.mock.sentinel.jidstr,
            strict=False
        )

        self.assertEqual(result, JID.fromstr())

    def test_parse_can_be_set_to_strict(self):
        with unittest.mock.patch("aioxmpp.structs.JID") as JID:
            t = xso.JID(strict=True)
            result = t.parse(unittest.mock.sentinel.jidstr)

        JID.fromstr.assert_called_with(
            unittest.mock.sentinel.jidstr,
            strict=True
        )

        self.assertEqual(result, JID.fromstr())

    def test_format(self):
        t = xso.JID()
        self.assertEqual(
            "ssa@ix.test/IX",
            t.format(structs.JID("ßA", "IX.test", "\u2168"))
        )

    def test_coerce_rejects_non_jids(self):
        t = xso.JID()
        types = [str, int, float, object]
        for type_ in types:
            with self.assertRaisesRegex(TypeError,
                                        "not a JID"):
                t.coerce(type_())

    def test_coerce_rejects_str_jids(self):
        t = xso.JID()
        with self.assertRaisesRegex(
                TypeError,
                "<class 'str'> object 'foo@bar' is not a JID"):
            t.coerce("foo@bar")

    def test_coerce_passes_jid(self):
        t = xso.JID()

        values = [
            structs.JID.fromstr("foo@bar.example"),
            structs.JID.fromstr("bar.example"),
            structs.JID.fromstr("foo@bar.example/baz"),
        ]

        for value in values:
            self.assertIs(
                value,
                t.coerce(value)
            )


class TestConnectionLocation(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.ConnectionLocation(),
            xso.AbstractCDataType)

    def test_parse_ipv6(self):
        t = xso.ConnectionLocation()
        self.assertEqual(
            (ipaddress.IPv6Address("fe80::"), 5222),
            t.parse("[fe80::]:5222")
        )

    def test_reject_non_integer_port_number(self):
        t = xso.ConnectionLocation()
        with self.assertRaises(ValueError):
            t.parse("[fe80::]:23.4")

    def test_reject_out_of_range_port_number(self):
        t = xso.ConnectionLocation()
        with self.assertRaises(ValueError):
            t.parse("[fe80::]:1000000")

    def test_reject_missing_colon(self):
        t = xso.ConnectionLocation()
        with self.assertRaises(ValueError):
            t.parse("foo.bar")

    def test_parse_ipv4(self):
        t = xso.ConnectionLocation()
        self.assertEqual(
            (ipaddress.IPv4Address("10.0.0.1"), 5223),
            t.parse("10.0.0.1:5223")
        )

    def test_parse_hostname(self):
        t = xso.ConnectionLocation()
        self.assertEqual(
            ("foo.bar.example", 5234),
            t.parse("foo.bar.example:5234")
        )

    def test_format_ipv6(self):
        t = xso.ConnectionLocation()
        self.assertEqual(
            "[fe80::]:5222",
            t.format((ipaddress.IPv6Address("fe80::"), 5222))
        )

    def test_format_ipv4(self):
        t = xso.ConnectionLocation()
        self.assertEqual(
            "10.0.0.1:1234",
            t.format((ipaddress.IPv4Address("10.0.0.1"), 1234))
        )

    def test_format_hostname(self):
        t = xso.ConnectionLocation()
        self.assertEqual(
            "foo.bar.baz:5234",
            t.format(("foo.bar.baz", 5234))
        )

    def test_coerce_rejects_non_2tuples(self):
        t = xso.ConnectionLocation()

        values = [
            ["foo", 1234],
            {"foo", 1234},
            ("foo", 1234, "bar")
        ]

        for value in values:
            with self.assertRaisesRegex(TypeError,
                                        "2-tuple required"):
                t.coerce(value)

    def test_coerce_parses_ip_addresses(self):
        t = xso.ConnectionLocation()

        value_pairs = [
            (("10.0.0.1", 1234), (ipaddress.IPv4Address("10.0.0.1"), 1234)),
            (("fe80::", 1234), (ipaddress.IPv6Address("fe80::"), 1234)),
            (("10.0.foobar", 1234), ("10.0.foobar", 1234)),
        ]

        for given, expected in value_pairs:
            self.assertEqual(
                expected,
                t.coerce(given)
            )

    def test_coerce_restricts_port_numbers(self):
        t = xso.ConnectionLocation()

        err_values = [
            ("foobar", -1),
            ("foobar", 65536),
        ]

        for err_value in err_values:
            with self.assertRaisesRegex(ValueError, "out of range"):
                t.coerce(err_value)

        ok_values = [
            ("foobar", 0),
            ("foobar", 65535),
        ]

        for ok_value in ok_values:
            self.assertEqual(
                ok_value,
                t.coerce(ok_value)
            )

    def test_coerce_requires_integral_number(self):
        t = xso.ConnectionLocation()

        values = [
            ("foobar", 1.2),
            ("foobar", "1"),
            ("foobar", decimal.Decimal("1")),
            ("foobar", fractions.Fraction(1, 1)),
        ]

        for value in values:
            with self.assertRaisesRegex(
                    TypeError,
                    "port number must be integral number"):
                t.coerce(value)


class TestLanguageTag(unittest.TestCase):
    def test_is_cdata_type(self):
        self.assertIsInstance(
            xso.LanguageTag(),
            xso.AbstractCDataType)

    def test_parse(self):
        t = xso.LanguageTag()
        self.assertEqual(
            structs.LanguageTag.fromstr("de-Latn-DE-1999"),
            t.parse("de-Latn-DE-1999")
        )

    def test_format(self):
        t = xso.LanguageTag()
        self.assertEqual(
            "de-Latn-DE-1999",
            t.format(structs.LanguageTag.fromstr("de-Latn-DE-1999"))
        )

    def test_coerce_passes_language_tags(self):
        t = xso.LanguageTag()
        tag = structs.LanguageTag.fromstr("foo")
        self.assertIs(
            tag,
            t.coerce(tag)
        )

    def test_coerce_rejects_non_language_tags(self):
        t = xso.LanguageTag()

        values = [
            1.2,
            decimal.Decimal("1"),
            fractions.Fraction(1, 1),
            [],
            (),
            1.,
            "foo",
        ]

        for value in values:
            with self.assertRaisesRegex(
                    TypeError,
                    "is not a LanguageTag"):
                t.coerce(value)


class TestTextChildMap(unittest.TestCase):
    def test_is_element_type(self):
        self.assertTrue(issubclass(
            xso.TextChildMap,
            xso.AbstractElementType
        ))

    def setUp(self):
        self.type_ = xso.TextChildMap(xso.AbstractTextChild)

    def tearDown(self):
        del self.type_

    def test_get_xso_types(self):
        self.assertCountEqual(
            self.type_.get_xso_types(),
            [xso.AbstractTextChild]
        )

    def test_unpack(self):
        text, lang = "foo", structs.LanguageTag.fromstr("en-gb")
        item = xso.AbstractTextChild(text, lang)
        self.assertEqual(
            (lang, text),
            self.type_.unpack(item)
        )

    def test_pack(self):
        text, lang = "foo", structs.LanguageTag.fromstr("en-gb")
        item = self.type_.pack((lang, text))

        self.assertEqual(item.text, text)
        self.assertEqual(item.lang, lang)


class TestEnumCDataType(unittest.TestCase):
    class SomeEnum(Enum):
        X = 1
        Y = 2
        Z = 3

    class SomeIntEnum(IntEnum):
        X = 1
        Y = 2
        Z = 3

    def test_is_cdata_type(self):
        self.assertTrue(issubclass(
            xso.EnumCDataType,
            xso.AbstractCDataType,
        ))

    def test_init_default(self):
        with self.assertRaises(TypeError):
            xso.EnumCDataType()

    def test_init_with_enum(self):
        e = xso.EnumCDataType(self.SomeEnum)
        self.assertIs(
            e.enum_class,
            self.SomeEnum
        )
        self.assertIsInstance(
            e.nested_type,
            xso.String,
        )

    def test_init_with_custom_nested_type(self):
        e = xso.EnumCDataType(
            self.SomeEnum,
            nested_type=unittest.mock.sentinel.nested_type
        )
        self.assertIs(
            e.enum_class,
            self.SomeEnum
        )
        self.assertIs(
            e.nested_type,
            unittest.mock.sentinel.nested_type,
        )

    def test_parse_uses_enum_and_nested_type(self):
        enum_class = unittest.mock.Mock()
        nested_type = unittest.mock.Mock()
        e = xso.EnumCDataType(enum_class, nested_type)

        result = e.parse(unittest.mock.sentinel.value)

        nested_type.parse.assert_called_with(
            unittest.mock.sentinel.value,
        )

        enum_class.assert_called_with(
            nested_type.parse(),
        )

        self.assertEqual(
            result,
            enum_class(),
        )

    def test_parse_works_with_actual_enum(self):
        e = xso.EnumCDataType(self.SomeEnum, xso.Integer())
        for enum_value in self.SomeEnum:
            self.assertEqual(
                e.parse(str(enum_value.value)),
                enum_value,
            )

    def test_format_uses_enum_value_and_nested_type(self):
        enum_class = unittest.mock.Mock()
        enum_value = unittest.mock.Mock()
        nested_type = unittest.mock.Mock()
        e = xso.EnumCDataType(enum_class, nested_type)

        result = e.format(enum_value)

        nested_type.format.assert_called_with(
            enum_value.value,
        )

        self.assertEqual(
            result,
            nested_type.format(),
        )

    def test_format_works_with_actual_enums(self):
        e = xso.EnumCDataType(self.SomeEnum, xso.Integer())
        for enum_value in self.SomeEnum:
            self.assertEqual(
                e.format(enum_value),
                str(enum_value.value),
            )

    def test_get_formatted_type_not_implemented(self):
        self.assertFalse(
            hasattr(xso.EnumCDataType, "get_formatted_type")
        )

    def test_get_xso_types_not_implemented(self):
        self.assertFalse(
            hasattr(xso.EnumCDataType, "get_xso_types")
        )

    def test_pass_Enum_values_through_coerce(self):
        e = xso.EnumCDataType(self.SomeEnum)
        for enum_value in self.SomeEnum:
            self.assertIs(enum_value, e.coerce(enum_value))

    def test_reject_non_Enum_values_on_coerce(self):
        wrong = [
            1,
            "10",
            10.2,
            object()
        ]

        e = xso.EnumCDataType(self.SomeEnum)

        for thing in wrong:
            with self.assertRaises(TypeError):
                e.coerce(thing)

    def test_try_to_coerce_if_allow_coerce_is_set(self):
        enum_class = unittest.mock.Mock()
        enum_class.return_value = unittest.mock.sentinel.wrapped
        nested_t = xso.Integer()
        t = xso.EnumCDataType(
            enum_class,
            nested_t,
            allow_coerce=True,
        )

        with contextlib.ExitStack() as stack:
            w = stack.enter_context(warnings.catch_warnings())
            coerce = stack.enter_context(
                unittest.mock.patch.object(nested_t, "coerce")
            )

            coerce.return_value = unittest.mock.sentinel.coerced
            result = t.coerce(unittest.mock.sentinel.value)

        enum_class.assert_called_with(
            unittest.mock.sentinel.coerced,
        )

        self.assertEqual(
            result,
            unittest.mock.sentinel.wrapped,
        )

        self.assertFalse(w)

    def test_value_error_propagates(self):
        exc = ValueError()

        enum_class = unittest.mock.Mock()
        enum_class.side_effect = exc
        e = xso.EnumCDataType(enum_class, xso.Integer(), allow_coerce=True)

        with self.assertRaises(ValueError) as ctx:
            e.coerce(1234)

        self.assertIs(ctx.exception, exc)

    def test_deprecate_coerce(self):
        enum_class = self.SomeEnum
        e = xso.EnumCDataType(
            enum_class,
            xso.Integer(),
            allow_coerce=True,
            deprecate_coerce=True,
        )

        with contextlib.ExitStack() as stack:
            warn = stack.enter_context(
                unittest.mock.patch(
                    "warnings.warn",
                )
            )

            result = e.coerce(1)

        warn.assert_called_with(
            "assignment of non-enum values to this descriptor is deprecated",
            DeprecationWarning,
            stacklevel=4
        )

        self.assertEqual(
            result,
            enum_class(1),
        )

    def test_deprecate_coerce_custom_stacklevel(self):
        enum_class = self.SomeEnum
        e = xso.EnumCDataType(
            enum_class,
            xso.Integer(),
            allow_coerce=True,
            deprecate_coerce=unittest.mock.sentinel.stacklevel,
        )

        with contextlib.ExitStack() as stack:
            warn = stack.enter_context(
                unittest.mock.patch(
                    "warnings.warn",
                )
            )

            result = e.coerce(1)

        warn.assert_called_with(
            "assignment of non-enum values to this descriptor is deprecated",
            DeprecationWarning,
            stacklevel=unittest.mock.sentinel.stacklevel
        )

        self.assertEqual(
            result,
            enum_class(1),
        )

    def test_deprecate_coerce_does_not_emit_warning_for_enum_value(self):
        enum_class = self.SomeEnum
        e = xso.EnumCDataType(
            enum_class,
            xso.Integer(),
            allow_coerce=True,
            deprecate_coerce=True,
        )

        value = enum_class.X

        with contextlib.ExitStack() as stack:
            warn = stack.enter_context(
                unittest.mock.patch(
                    "warnings.warn",
                )
            )

            result = e.coerce(value)

        self.assertFalse(warn.mock_calls)

        self.assertIs(
            value,
            result,
        )

    def test_accept_unknown_by_default(self):
        enum_class = self.SomeEnum
        e = xso.EnumCDataType(
            enum_class,
            xso.Integer(),
        )

        value = e.coerce(xso.Unknown(10))
        self.assertIsInstance(value, xso.Unknown)
        self.assertEqual(xso.Unknown(10), value)

    def test_accept_unknown_can_be_turned_off(self):
        enum_class = self.SomeEnum
        e = xso.EnumCDataType(
            enum_class,
            xso.Integer(),
            accept_unknown=False,
        )

        with self.assertRaisesRegex(
                TypeError,
                r"not a valid .* value: <Unknown: 10>"):
            e.coerce(xso.Unknown(10))

    def test_allow_unknown_by_default(self):
        enum_class = self.SomeEnum
        e = xso.EnumCDataType(
            enum_class,
            xso.Integer(),
        )

        value = e.parse("10")
        self.assertIsInstance(value, xso.Unknown)
        self.assertEqual(xso.Unknown(10), value)

    def test_allow_unknown_can_be_turned_off(self):
        enum_class = self.SomeEnum
        e = xso.EnumCDataType(
            enum_class,
            xso.Integer(),
            allow_unknown=False,
        )

        with self.assertRaisesRegex(
                ValueError,
                r"10 is not a valid SomeEnum"):
            e.parse(10)

    def test_format_works_with_unknown(self):
        enum_class = self.SomeEnum
        e = xso.EnumCDataType(
            enum_class,
            xso.Integer(),
        )

        self.assertEqual(
            e.format(xso.Unknown(10)),
            "10",
        )

    def test_reject_pass_unknown_without_allow_unknown(self):
        enum_class = self.SomeEnum

        with self.assertRaisesRegex(
                ValueError,
                r"pass_unknown requires allow_unknown and accept_unknown"):
            xso.EnumCDataType(
                enum_class,
                xso.Integer(),
                allow_unknown=False,
                pass_unknown=True,
            )

    def test_reject_pass_unknown_without_accept_unknown(self):
        enum_class = self.SomeEnum

        with self.assertRaisesRegex(
                ValueError,
                r"pass_unknown requires allow_unknown and accept_unknown"):
            xso.EnumCDataType(
                enum_class,
                xso.Integer(),
                allow_unknown=False,
                pass_unknown=True,
            )

    def test_coerce_passes_non_members_with_pass_unknown(self):
        t = xso.EnumCDataType(
            self.SomeIntEnum,
            xso.Integer(),
            pass_unknown=True,
        )

        v = t.coerce(10)
        self.assertFalse(isinstance(v, IntEnum))

    def test_coerce_passes_value_to_nested_type_coerce_with_pass_unknown(self):
        nested_t = xso.Integer()
        t = xso.EnumCDataType(
            self.SomeIntEnum,
            nested_t,
            pass_unknown=True,
        )

        with unittest.mock.patch.object(nested_t, "coerce") as coerce:
            coerce.return_value = unittest.mock.sentinel.coerced

            v = t.coerce(unittest.mock.sentinel.value)

        self.assertEqual(v, unittest.mock.sentinel.coerced)

    def test_coerce_converts_to_enum_members_if_allow_coerce_is_set(self):
        nested_t = xso.Integer()
        t = xso.EnumCDataType(
            self.SomeIntEnum,
            nested_t,
            allow_coerce=True,
            pass_unknown=True,
        )

        with unittest.mock.patch.object(nested_t, "coerce") as coerce:
            coerce.return_value = 2

            v = t.coerce(unittest.mock.sentinel.value)

        self.assertEqual(v, 2)
        self.assertIsInstance(v, self.SomeIntEnum)

    def test_coerce_does_not_convert_to_enum_members_if_allow_coerce_unset(self):  # NOQA
        nested_t = xso.Integer()
        t = xso.EnumCDataType(
            self.SomeIntEnum,
            nested_t,
            allow_coerce=False,
            pass_unknown=True,
        )

        with unittest.mock.patch.object(nested_t, "coerce") as coerce:
            coerce.return_value = 2

            v = t.coerce(unittest.mock.sentinel.value)

        self.assertEqual(v, 2)
        self.assertFalse(isinstance(v, self.SomeIntEnum))

    def test_coerce_passes_non_members_with_pass_unknown_and_allow_coerce(self):
        t = xso.EnumCDataType(
            self.SomeIntEnum,
            xso.Integer(),
            allow_coerce=True,
            pass_unknown=True,
        )

        v = t.coerce(10)
        self.assertFalse(isinstance(v, IntEnum))

    def test_parse_passes_unwrapped_value_if_pass_unknown(self):
        t = xso.EnumCDataType(
            self.SomeIntEnum,
            xso.Integer(),
            pass_unknown=True,
        )

        v = t.parse("10")
        self.assertEqual(v, 10)
        self.assertFalse(isinstance(v, xso.Unknown))

    def test_format_works_with_unwrapped_unknowns_if_pass_unknown(self):
        t = xso.EnumCDataType(
            self.SomeIntEnum,
            xso.Integer(),
            pass_unknown=True,
        )

        v = t.format(10)
        self.assertEqual(v, "10")


class TestEnumElementType(unittest.TestCase):
    class SomeEnum(Enum):
        X = 1
        Y = 2
        Z = 3

    class FancyType(xso.AbstractElementType):
        def get_xso_types(self):
            raise NotImplementedError

        def pack(self, obj):
            return obj

        def unpack(self, obj):
            return obj

    def test_is_element_type(self):
        self.assertTrue(issubclass(
            xso.EnumElementType,
            xso.AbstractElementType,
        ))

    def test_init_default(self):
        with self.assertRaises(TypeError):
            xso.EnumElementType()

    def test_init_with_enum(self):
        with self.assertRaises(TypeError):
            xso.EnumElementType(self.SomeEnum)

    def test_init_with_custom_nested_type(self):
        e = xso.EnumElementType(
            self.SomeEnum,
            nested_type=unittest.mock.sentinel.nested_type
        )
        self.assertIs(
            e.enum_class,
            self.SomeEnum
        )
        self.assertIs(
            e.nested_type,
            unittest.mock.sentinel.nested_type,
        )

    def test_unpack_uses_enum_and_nested_type(self):
        enum_class = unittest.mock.Mock()
        nested_type = unittest.mock.Mock()
        e = xso.EnumElementType(enum_class, nested_type)

        result = e.unpack(unittest.mock.sentinel.value)

        nested_type.unpack.assert_called_with(
            unittest.mock.sentinel.value,
        )

        enum_class.assert_called_with(
            nested_type.unpack(),
        )

        self.assertEqual(
            result,
            enum_class(),
        )

    def test_unpack_works_with_actual_enum(self):
        e = xso.EnumElementType(self.SomeEnum, self.FancyType())
        for enum_value in self.SomeEnum:
            self.assertEqual(
                e.unpack(enum_value.value),
                enum_value,
            )

    def test_pack_uses_enum_value_and_nested_type(self):
        enum_class = unittest.mock.Mock()
        enum_value = unittest.mock.Mock()
        nested_type = unittest.mock.Mock()
        e = xso.EnumElementType(enum_class, nested_type)

        result = e.pack(enum_value)

        nested_type.pack.assert_called_with(
            enum_value.value,
        )

        self.assertEqual(
            result,
            nested_type.pack(),
        )

    def test_pack_works_with_actual_enums(self):
        e = xso.EnumElementType(self.SomeEnum, self.FancyType())
        for enum_value in self.SomeEnum:
            self.assertEqual(
                e.pack(enum_value),
                enum_value.value,
            )

    def test_get_xso_types_delegates_to_nested_type(self):
        nested_type = unittest.mock.Mock()
        e = xso.EnumElementType(
            unittest.mock.sentinel.enum_class,
            nested_type,
        )

        result = e.get_xso_types()
        nested_type.get_xso_types.assert_called_with()
        self.assertEqual(
            result,
            nested_type.get_xso_types(),
        )

    def test_pass_Enum_values_through_coerce(self):
        e = xso.EnumElementType(self.SomeEnum, self.FancyType())
        for enum_value in self.SomeEnum:
            self.assertIs(enum_value, e.coerce(enum_value))

    def test_reject_non_Enum_values_on_coerce(self):
        wrong = [
            1,
            "10",
            10.2,
            object()
        ]

        e = xso.EnumElementType(self.SomeEnum, self.FancyType())

        for thing in wrong:
            with self.assertRaises(TypeError):
                e.coerce(thing)

    def test_try_to_coerce_if_allow_coerce_is_set(self):
        enum_class = unittest.mock.Mock()
        e = xso.EnumElementType(
            enum_class,
            self.FancyType(),
            allow_coerce=True,
        )

        with warnings.catch_warnings() as w:
            result = e.coerce(unittest.mock.sentinel.value)

        enum_class.assert_called_with(
            unittest.mock.sentinel.value,
        )

        self.assertEqual(
            result,
            enum_class(),
        )

        self.assertFalse(w)

    def test_value_error_propagates(self):
        exc = ValueError()

        enum_class = unittest.mock.Mock()
        enum_class.side_effect = exc
        e = xso.EnumElementType(enum_class, self.FancyType(),
                                allow_coerce=True)

        with self.assertRaises(ValueError) as ctx:
            e.coerce(unittest.mock.sentinel.value)

        self.assertIs(ctx.exception, exc)

    def test_deprecate_coerce(self):
        enum_class = self.SomeEnum
        e = xso.EnumElementType(
            enum_class,
            self.FancyType(),
            allow_coerce=True,
            deprecate_coerce=True,
        )

        with contextlib.ExitStack() as stack:
            warn = stack.enter_context(
                unittest.mock.patch(
                    "warnings.warn",
                )
            )

            result = e.coerce(1)

        warn.assert_called_with(
            "assignment of non-enum values to this descriptor is deprecated",
            DeprecationWarning,
            stacklevel=4
        )

        self.assertEqual(
            result,
            enum_class(1),
        )

    def test_deprecate_coerce_custom_stacklevel(self):
        enum_class = self.SomeEnum
        e = xso.EnumElementType(
            enum_class,
            self.FancyType(),
            allow_coerce=True,
            deprecate_coerce=unittest.mock.sentinel.stacklevel,
        )

        with contextlib.ExitStack() as stack:
            warn = stack.enter_context(
                unittest.mock.patch(
                    "warnings.warn",
                )
            )

            result = e.coerce(1)

        warn.assert_called_with(
            "assignment of non-enum values to this descriptor is deprecated",
            DeprecationWarning,
            stacklevel=unittest.mock.sentinel.stacklevel
        )

        self.assertEqual(
            result,
            enum_class(1),
        )

    def test_deprecate_coerce_does_not_emit_warning_for_enum_value(self):
        enum_class = self.SomeEnum
        e = xso.EnumElementType(
            enum_class,
            self.FancyType(),
            allow_coerce=True,
            deprecate_coerce=True,
        )

        value = enum_class.X

        with contextlib.ExitStack() as stack:
            warn = stack.enter_context(
                unittest.mock.patch(
                    "warnings.warn",
                )
            )

            result = e.coerce(value)

        self.assertFalse(warn.mock_calls)

        self.assertIs(
            value,
            result,
        )

    def test_accept_unknown_by_default(self):
        enum_class = self.SomeEnum
        e = xso.EnumElementType(
            enum_class,
            xso.Integer(),
        )

        value = e.coerce(xso.Unknown(10))
        self.assertIsInstance(value, xso.Unknown)
        self.assertEqual(xso.Unknown(10), value)

    def test_accept_unknown_can_be_turned_off(self):
        enum_class = self.SomeEnum
        e = xso.EnumElementType(
            enum_class,
            xso.Integer(),
            accept_unknown=False,
        )

        with self.assertRaisesRegex(
                TypeError,
                r"not a valid .* value: <Unknown: 10>"):
            e.coerce(xso.Unknown(10))

    def test_allow_unknown_by_default(self):
        enum_class = self.SomeEnum
        e = xso.EnumElementType(
            enum_class,
            self.FancyType(),
        )

        value = e.unpack(10)
        self.assertIsInstance(value, xso.Unknown)
        self.assertEqual(xso.Unknown(10), value)

    def test_allow_unknown_can_be_turned_off(self):
        enum_class = self.SomeEnum
        e = xso.EnumElementType(
            enum_class,
            self.FancyType(),
            allow_unknown=False,
        )

        with self.assertRaisesRegex(
                ValueError,
                r"10 is not a valid SomeEnum"):
            e.unpack(10)

    def test_format_works_with_unknown(self):
        enum_class = self.SomeEnum
        e = xso.EnumElementType(
            enum_class,
            self.FancyType(),
        )

        self.assertEqual(
            e.pack(xso.Unknown(10)),
            10,
        )


class TestAbstractValidator(unittest.TestCase):
    def test_is_abstract(self):
        self.assertIsInstance(
            xso.AbstractValidator,
            abc.ABCMeta)
        with self.assertRaises(TypeError):
            xso.AbstractValidator()

    def test_validate_calls_validate_detailed(self):
        class FakeSubclass(xso.AbstractValidator):
            def validate_detailed(self, value):
                pass

        instance = FakeSubclass()
        obj = object()
        with unittest.mock.patch.object(instance, "validate_detailed") \
                 as validate_detailed:
            instance.validate(obj)

        self.assertSequenceEqual(
            [
                unittest.mock.call(obj),
                unittest.mock.call().__bool__(),
            ],
            validate_detailed.mock_calls
        )

    def test_validate_calls_validate_detailed_and_inverts_result(self):
        class FakeSubclass(xso.AbstractValidator):
            def validate_detailed(self, value):
                return []

        instance = FakeSubclass()
        obj = object()
        self.assertTrue(instance.validate(obj))


class TestRestrictToSet(unittest.TestCase):
    def test_is_abstract_validator(self):
        self.assertIsInstance(
            xso.RestrictToSet([]),
            xso.AbstractValidator)

    def test_validate(self):
        t = xso.RestrictToSet({"foo", "bar"})
        self.assertTrue(t.validate("foo"))
        self.assertTrue(t.validate("bar"))
        self.assertFalse(t.validate("baz"))


class TestNmtoken(unittest.TestCase):
    def test_is_abstract_validator(self):
        self.assertIsInstance(
            xso.RestrictToSet([]),
            xso.AbstractValidator)

    def _test_samples(self, t, samples, group):
        for sample in samples:
            self.assertTrue(
                t.validate(sample),
                "\\u{:04x} is supposed to be in {}".format(ord(sample), group)
            )

    def test_validate(self):
        t = xso.Nmtoken()
        # ok, testing this sucks hard. we’ll do some hand-waiving tests
        # guarding against the most important characters which must not occur.

        self.assertTrue(t.validate("foobar"))
        self.assertTrue(t.validate("foo:bar"))
        self.assertTrue(t.validate("foo-bar"))
        self.assertTrue(t.validate("foo.bar"))
        self.assertTrue(t.validate("foo_bar"))
        self.assertTrue(t.validate("."))
        self.assertTrue(t.validate(":"))
        self.assertTrue(t.validate("_"))
        self.assertTrue(t.validate("."))

        self.assertFalse(t.validate("\uf901"))
        self.assertFalse(t.validate("\ufffd"))
        self.assertFalse(t.validate("\u20dd"))
        self.assertFalse(t.validate(">"))
        self.assertFalse(t.validate("<"))
        self.assertFalse(t.validate("&"))

    def test_validate_base_char(self):
        self._test_samples(
            xso.Nmtoken(),
            "\u0041"
            "\u0061"
            "\u00c0"
            "\u00d8"
            "\u00F8"
            "\u0100"
            "\u0134"
            "\u0141"
            "\u014A"
            "\u0180"
            "\u01CD"
            "\u01F4"
            "\u01FA"
            "\u0250"
            "\u02BB"
            "\u0386"
            "\u0388"
            "\u038C"
            "\u038E"
            "\u03A3"
            "\u03D0"
            "\u03DA"
            "\u03DC"
            "\u03DE"
            "\u03E0"
            "\u03E2"
            "\u0401"
            "\u040E"
            "\u0451"
            "\u045E"
            "\u0490"
            "\u04C7"
            "\u04CB"
            "\u04D0"
            "\u04EE"
            "\u04F8"
            "\u0531"
            "\u0559"
            "\u0561"
            "\u05D0"
            "\u05F0"
            "\u0621"
            "\u0641"
            "\u0671"
            "\u06BA"
            "\u06C0"
            "\u06D0"
            "\u06D5"
            "\u06E5"
            "\u0905"
            "\u093D"
            "\u0958"
            "\u0985"
            "\u098F"
            "\u0993"
            "\u09AA"
            "\u09B2"
            "\u09B6"
            "\u09DC"
            "\u09DF"
            "\u09F0"
            "\u0A05"
            "\u0A0F"
            "\u0A13"
            "\u0A2A"
            "\u0A32"
            "\u0A35"
            "\u0A38"
            "\u0A59"
            "\u0A5E"
            "\u0A72"
            "\u0A85"
            "\u0A8D"
            "\u0A8F"
            "\u0A93"
            "\u0AAA"
            "\u0AB2"
            "\u0AB5"
            "\u0ABD"
            "\u0AE0"
            "\u0B05"
            "\u0B0F"
            "\u0B13"
            "\u0B2A"
            "\u0B32"
            "\u0B36"
            "\u0B3D"
            "\u0B5C"
            "\u0B5F"
            "\u0B85"
            "\u0B8E"
            "\u0B92"
            "\u0B99"
            "\u0B9C"
            "\u0B9E"
            "\u0BA3"
            "\u0BA8"
            "\u0BAE"
            "\u0BB7"
            "\u0C05"
            "\u0C0E"
            "\u0C12"
            "\u0C2A"
            "\u0C35"
            "\u0C60"
            "\u0C85"
            "\u0C8E"
            "\u0C92"
            "\u0CAA"
            "\u0CB5"
            "\u0CDE"
            "\u0CE0"
            "\u0D05"
            "\u0D0E"
            "\u0D12"
            "\u0D2A"
            "\u0D60"
            "\u0E01"
            "\u0E30"
            "\u0E32"
            "\u0E40"
            "\u0E81"
            "\u0E84"
            "\u0E87"
            "\u0E8A"
            "\u0E8D"
            "\u0E94"
            "\u0E99"
            "\u0EA1"
            "\u0EA5"
            "\u0EA7"
            "\u0EAA"
            "\u0EAD"
            "\u0EB0"
            "\u0EB2"
            "\u0EBD"
            "\u0EC0"
            "\u0F40"
            "\u0F49"
            "\u10A0"
            "\u10D0"
            "\u1100"
            "\u1102"
            "\u1105"
            "\u1109"
            "\u110B"
            "\u110E"
            "\u113C"
            "\u113E"
            "\u1140"
            "\u114C"
            "\u114E"
            "\u1150"
            "\u1154"
            "\u1159"
            "\u115F"
            "\u1163"
            "\u1165"
            "\u1167"
            "\u1169"
            "\u116D"
            "\u1172"
            "\u1175"
            "\u119E"
            "\u11A8"
            "\u11AB"
            "\u11AE"
            "\u11B7"
            "\u11BA"
            "\u11BC"
            "\u11EB"
            "\u11F0"
            "\u11F9"
            "\u1E00"
            "\u1EA0"
            "\u1F00"
            "\u1F18"
            "\u1F20"
            "\u1F48"
            "\u1F50"
            "\u1F59"
            "\u1F5B"
            "\u1F5D"
            "\u1F5F"
            "\u1F80"
            "\u1FB6"
            "\u1FBE"
            "\u1FC2"
            "\u1FC6"
            "\u1FD0"
            "\u1FD6"
            "\u1FE0"
            "\u1FF2"
            "\u1FF6"
            "\u2126"
            "\u212A"
            "\u212E"  # deliberately excluded
            "\u2180"
            "\u3041"
            "\u30A1"
            "\u3105"
            "\uAC00",
            "BaseChar"
        )

    def test_validate_ideographic(self):
        self._test_samples(
            xso.Nmtoken(),
            "\u4E00"
            "\u3007"
            "\u3021",
            "Ideographic"
        )

    def test_validate_combining(self):
        self._test_samples(
            xso.Nmtoken(),
            "\u0300"
            "\u0360"
            "\u0483"
            "\u0591"
            "\u05A3"
            "\u05BB"
            "\u05BF"
            "\u05C1"
            "\u05C4"
            "\u064B"
            "\u0670"
            "\u06D6"
            "\u06DD"
            "\u06E0"
            "\u06E7"
            "\u06EA"
            "\u0901"
            "\u093C"
            "\u093E"
            "\u094D"
            "\u0951"
            "\u0962"
            "\u0981"
            "\u09BC"
            "\u09BE"
            "\u09BF"
            "\u09C0"
            "\u09C7"
            "\u09CB"
            "\u09D7"
            "\u09E2"
            "\u0A02"
            "\u0A3C"
            "\u0A3E"
            "\u0A3F"
            "\u0A40"
            "\u0A47"
            "\u0A4B"
            "\u0A70"
            "\u0A81"
            "\u0ABC"
            "\u0ABE"
            "\u0AC7"
            "\u0ACB"
            "\u0B01"
            "\u0B3C"
            "\u0B3E"
            "\u0B47"
            "\u0B4B"
            "\u0B56"
            "\u0B82"
            "\u0BBE"
            "\u0BC6"
            "\u0BCA"
            "\u0BD7"
            "\u0C01"
            "\u0C3E"
            "\u0C46"
            "\u0C4A"
            "\u0C55"
            "\u0C82"
            "\u0CBE"
            "\u0CC6"
            "\u0CCA"
            "\u0CD5"
            "\u0D02"
            "\u0D3E"
            "\u0D46"
            "\u0D4A"
            "\u0D57"
            "\u0E31"
            "\u0E34"
            "\u0E47"
            "\u0EB1"
            "\u0EB4"
            "\u0EBB"
            "\u0EC8"
            "\u0F18"
            "\u0F35"
            "\u0F37"
            "\u0F39"
            "\u0F3E"
            "\u0F3F"
            "\u0F71"
            "\u0F86"
            "\u0F90"
            "\u0F97"
            "\u0F99"
            "\u0FB1"
            "\u0FB9"
            "\u20D0"
            "\u20E1"
            "\u302A"
            "\u3099"
            "\u309A",
            "CombiningChar"
        )

    def test_validate_digit(self):
        self._test_samples(
            xso.Nmtoken(),
            "\u0030"
            "\u0660"
            "\u06F0"
            "\u0966"
            "\u09E6"
            "\u0A66"
            "\u0AE6"
            "\u0B66"
            "\u0BE7"
            "\u0C66"
            "\u0CE6"
            "\u0D66"
            "\u0E50"
            "\u0ED0"
            "\u0F20",
            "Digit"
        )

    def test_validate_extender(self):
        self._test_samples(
            xso.Nmtoken(),
            "\u00B7"
            "\u02D0"
            "\u02D1"
            "\u0387"
            "\u0640"
            "\u0E46"
            "\u0EC6"
            "\u3005"
            "\u3031"
            "\u309D"
            "\u30FC",
            "Extender"
        )


class TestIsInstance(unittest.TestCase):
    def test_is_abstract_validator(self):
        self.assertTrue(issubclass(
            xso.IsInstance,
            xso.AbstractValidator
        ))

    def test_validate(self):
        v = xso.IsInstance((str, bytes))
        self.assertTrue(
            v.validate("abc")
        )
        self.assertTrue(
            v.validate(b"abc")
        )
        self.assertFalse(
            v.validate(1)
        )

    def test_list_of_classes_is_shared(self):
        classes = []
        v = xso.IsInstance(classes)

        self.assertFalse(
            v.validate("str")
        )

        classes.append(str)

        self.assertTrue(
            v.validate("str")
        )


class TestNumericRange(unittest.TestCase):
    def test_is_abstract_validator(self):
        self.assertTrue(issubclass(
            xso.NumericRange,
            xso.AbstractValidator
        ))

    def test_validate_ok(self):
        v = xso.NumericRange(min_=10, max_=20)
        for i in range(10, 21):
            self.assertTrue(
                v.validate(i),
            )

    def test_validate_detailed_out_of_bounds(self):
        v = xso.NumericRange(min_=10, max_=20)
        for i in range(0, 10):
            self.assertFalse(
                v.validate(i),
            )

    def test_validate_detailed_too_small(self):
        v = xso.NumericRange(min_=10)
        for i in range(0, 10):
            self.assertFalse(
                v.validate(i),
            )

    def test_validate_detailed_too_large(self):
        v = xso.NumericRange(max_=-1)
        for i in range(0, 10):
            self.assertFalse(
                v.validate(i),
            )


class TestEnumType(unittest.TestCase):
    def test_instanciates_EnumCDataType_by_default_and_passes_kwargs(self):
        with contextlib.ExitStack() as stack:
            EnumCDataType = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.types.EnumCDataType")
            )

            result = xso.EnumType(
                unittest.mock.sentinel.enum_class,
                foo=unittest.mock.sentinel.foo,
                bar=unittest.mock.sentinel.bar,
            )

        EnumCDataType.assert_called_once_with(
            unittest.mock.sentinel.enum_class,
            foo=unittest.mock.sentinel.foo,
            bar=unittest.mock.sentinel.bar,
        )

        self.assertEqual(result, EnumCDataType())

    def test_instanciates_EnumCDataType_for_AbstractCDataType(self):
        m = unittest.mock.Mock(
            spec=xso.AbstractCDataType
        )

        with contextlib.ExitStack() as stack:
            EnumCDataType = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.types.EnumCDataType")
            )

            result = xso.EnumType(
                unittest.mock.sentinel.enum_class,
                m,
                foo=unittest.mock.sentinel.foo,
                bar=unittest.mock.sentinel.bar,
            )

        EnumCDataType.assert_called_once_with(
            unittest.mock.sentinel.enum_class,
            m,
            foo=unittest.mock.sentinel.foo,
            bar=unittest.mock.sentinel.bar,
        )

        self.assertEqual(result, EnumCDataType())

    def test_instanciates_EnumCDataType_for_AbstractElementType(self):
        m = unittest.mock.Mock(
            spec=xso.AbstractElementType
        )

        with contextlib.ExitStack() as stack:
            EnumElementType = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.types.EnumElementType")
            )

            result = xso.EnumType(
                unittest.mock.sentinel.enum_class,
                m,
                foo=unittest.mock.sentinel.foo,
                bar=unittest.mock.sentinel.bar,
            )

        EnumElementType.assert_called_once_with(
            unittest.mock.sentinel.enum_class,
            m,
            foo=unittest.mock.sentinel.foo,
            bar=unittest.mock.sentinel.bar,
        )

        self.assertEqual(result, EnumElementType())



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import collections
import copy
import functools
import logging
import os
import tempfile

import aioxmpp.callbacks
import aioxmpp.disco as disco
import aioxmpp.service
import aioxmpp.utils
import aioxmpp.xml
import aioxmpp.xso

from aioxmpp.utils import namespaces

from . import caps115, caps390


logger = logging.getLogger("aioxmpp.entitycaps")


class Cache:
    """
    This provides a two-level cache for entity capabilities information. The
    idea is to have a trusted database, e.g. installed system-wide or shipped
    with :mod:`aioxmpp` and in addition a user-level database which is
    automatically filled with hashes which have been found by the
    :class:`Service`.

    The trusted database is taken as read-only and overrides the user-collected
    database. When a hash is in both databases, it is removed from the
    user-collected database (to save space).

    In addition to serving the databases, it provides deduplication for queries
    by holding a cache of futures looking up the same hash.

    Database management (user API):

    .. automethod:: set_system_db_path

    .. automethod:: set_user_db_path

    Queries (API intended for :class:`Service`):

    .. automethod:: create_query_future

    .. automethod:: lookup_in_database

    .. automethod:: lookup
    """

    def __init__(self):
        self._lookup_cache = {}
        self._memory_overlay = {}
        self._system_db_path = None
        self._user_db_path = None

    def _erase_future(self, key, fut):
        try:
            existing = self._lookup_cache[key]
        except KeyError:
            pass
        else:
            if existing is fut:
                del self._lookup_cache[key]

    def set_system_db_path(self, path):
        self._system_db_path = path

    def set_user_db_path(self, path):
        self._user_db_path = path

    def lookup_in_database(self, key):
        try:
            result = self._memory_overlay[key]
        except KeyError:
            pass
        else:
            logger.debug("memory cache hit: %s", key)
            return result

        key_path = key.path

        if self._system_db_path is not None:
            try:
                f = (
                    self._system_db_path / key_path
                ).open("rb")
            except OSError:
                pass
            else:
                logger.debug("system db hit: %s", key)
                with f:
                    return aioxmpp.xml.read_single_xso(f, disco.xso.InfoQuery)

        if self._user_db_path is not None:
            try:
                f = (
                    self._user_db_path / key_path
                ).open("rb")
            except OSError:
                pass
            else:
                logger.debug("user db hit: %s", key)
                with f:
                    return aioxmpp.xml.read_single_xso(f, disco.xso.InfoQuery)

        raise KeyError(key)

    @asyncio.coroutine
    def lookup(self, key):
        """
        Look up the given `node` URL using the given `hash_` first in the
        database and then by waiting on the futures created with
        :meth:`create_query_future` for that node URL and hash.

        If the hash is not in the database, :meth:`lookup` iterates as long as
        there are pending futures for the given `hash_` and `node`. If there
        are no pending futures, :class:`KeyError` is raised. If a future raises
        a :class:`ValueError`, it is ignored. If the future returns a value, it
        is used as the result.
        """
        try:
            result = self.lookup_in_database(key)
        except KeyError:
            pass
        else:
            return result

        while True:
            fut = self._lookup_cache[key]
            try:
                result = yield from fut
            except ValueError:
                continue
            else:
                return result

    def create_query_future(self, key):
        """
        Create and return a :class:`asyncio.Future` for the given `hash_`
        function and `node` URL. The future is referenced internally and used
        by any calls to :meth:`lookup` which are made while the future is
        pending. The future is removed from the internal storage automatically
        when a result or exception is set for it.

        This allows for deduplication of queries for the same hash.
        """
        fut = asyncio.Future()
        fut.add_done_callback(
            functools.partial(self._erase_future, key)
        )
        self._lookup_cache[key] = fut
        return fut

    def add_cache_entry(self, key, entry):
        """
        Add the given `entry` (which must be a :class:`~.disco.xso.InfoQuery`
        instance) to the user-level database keyed with the hash function type
        `hash_` and the `node` URL. The `entry` is **not** validated to
        actually map to `node` with the given `hash_` function, it is expected
        that the caller perfoms the validation.
        """
        copied_entry = copy.copy(entry)
        self._memory_overlay[key] = copied_entry
        if self._user_db_path is not None:
            asyncio.ensure_future(asyncio.get_event_loop().run_in_executor(
                None,
                writeback,
                self._user_db_path / key.path,
                entry.captured_events))


class EntityCapsService(aioxmpp.service.Service):
    """
    Make use and provide service discovery information in presence broadcasts.

    This service implements :xep:`0115` and :xep:`0390`, transparently.
    Besides loading the service, no interaction is required to get some of
    the benefits of :xep:`0115` and :xep:`0390`.

    Two additional things need to be done by users to get full support and
    performance:

    1. To make sure that peers are always up-to-date with the current
       capabilities, it is required that users listen on the
       :meth:`on_ver_changed` signal and re-emit their current presence when it
       fires.

       .. note::

           Keeping peers up-to-date is a MUST in :xep:`390`.

       The service takes care of attaching capabilities information on the
       outgoing stanza, using a stanza filter.

       .. warning::

           :meth:`on_ver_changed` may be emitted at a considerable rate when
           services are loaded or certain features (such as PEP-based services)
           are configured. It is up to the application to limit the rate at
           which presences are sent for the sole purpose of updating peers with
           new capability information.

    2. Users should use a process-wide :class:`Cache` instance and assign it to
       the :attr:`cache` of each :class:`.entitycaps.Service` they use. This
       improves performance by sharing (verified) hashes among :class:`Service`
       instances.

       In addition, the hashes should be saved and restored on shutdown/start
       of the process. See the :class:`Cache` for details.

    .. signal:: on_ver_changed

       The signal emits whenever the Capability Hashset of the local client
       changes. This happens when the set of features or identities announced
       in the :class:`.DiscoServer` changes.

    .. autoattribute:: cache

    .. autoattribute:: xep115_support

    .. autoattribute:: xep390_support

    .. versionchanged:: 0.8

       This class was formerly known as :class:`aioxmpp.entitycaps.Service`. It
       is still available under that name, but the alias will be removed in
       1.0.

    .. versionchanged:: 0.9

        Support for :xep:`390` was added.

    """

    ORDER_AFTER = {
        disco.DiscoClient,
        disco.DiscoServer,
    }

    NODE = "http://aioxmpp.zombofant.net/"

    on_ver_changed = aioxmpp.callbacks.Signal()

    def __init__(self, node, **kwargs):
        super().__init__(node, **kwargs)

        self.__current_keys = {}
        self._cache = Cache()

        self.disco_server = self.dependencies[disco.DiscoServer]
        self.disco_client = self.dependencies[disco.DiscoClient]

        self.__115 = caps115.Implementation(self.NODE)
        self.__390 = caps390.Implementation(
            aioxmpp.hashes.default_hash_algorithms
        )

        self.__active_hashsets = []
        self.__key_users = collections.Counter()

    @property
    def xep115_support(self):
        """
        Boolean to control whether :xep:`115` support is enabled or not.

        Defaults to :data:`True`.

        If set to false, inbound :xep:`115` capabilities will not be processed
        and no :xep:`115` capabilities will be emitted.

        .. note::

            At some point, this will default to :data:`False` to save
            bandwidth. The exact release depends on the adoption of :xep:`390`
            and will be announced in time. If you depend on :xep:`115` support,
            set this boolean to :data:`True`.

            The attribute itself will not be removed until :xep:`115` support
            is removed from :mod:`aioxmpp` entirely, which is unlikely to
            happen any time soon.

        .. versionadded:: 0.9
        """

        return self._xep115_feature.enabled

    @xep115_support.setter
    def xep115_support(self, value):
        self._xep115_feature.enabled = value

    @property
    def xep390_support(self):
        """
        Boolean to control whether :xep:`390` support is enabled or not.

        Defaults to :data:`True`.

        If set to false, inbound :xep:`390` Capability Hash Sets will not be
        processed and no Capability Hash Sets or Capability Nodes will be
        generated.

        The hash algortihms used for generating Capability Hash Sets are those
        from :data:`aioxmpp.hashes.default_hash_algorithms`.
        """
        return self._xep390_feature.enabled

    @xep390_support.setter
    def xep390_support(self, value):
        self._xep390_feature.enabled = value

    @property
    def cache(self):
        """
        The :class:`Cache` instance used for this :class:`Service`. Deleting
        this attribute will automatically create a new :class:`Cache` instance.

        The attribute can be used to share a single :class:`Cache` among
        multiple :class:`Service` instances.
        """
        return self._cache

    @cache.setter
    def cache(self, v):
        self._cache = v

    @cache.deleter
    def cache(self):
        self._cache = Cache()

    @aioxmpp.service.depsignal(
        disco.DiscoServer,
        "on_info_changed")
    def _info_changed(self):
        self.logger.debug("info changed, scheduling re-calculation of version")
        asyncio.get_event_loop().call_soon(
            self.update_hash
        )

    @asyncio.coroutine
    def _shutdown(self):
        for group in self.__current_keys.values():
            for key in group:
                self.disco_server.unmount_node(key.node)

    @asyncio.coroutine
    def query_and_cache(self, jid, key, fut):
        data = yield from self.disco_client.query_info(
            jid,
            node=key.node,
            require_fresh=True,
            no_cache=True,  # the caps node is never queried by apps
        )

        try:
            if key.verify(data):
                self.cache.add_cache_entry(key, data)
                fut.set_result(data)
            else:
                raise ValueError("hash mismatch")
        except ValueError as exc:
            fut.set_exception(exc)

        return data

    @asyncio.coroutine
    def lookup_info(self, jid, keys):
        for key in keys:
            try:
                info = yield from self.cache.lookup(key)
            except KeyError:
                continue

            self.logger.debug("found %s in cache", key)
            return info

        first_key = keys[0]
        self.logger.debug("using key %s to query peer", first_key)
        fut = self.cache.create_query_future(first_key)
        info = yield from self.query_and_cache(
            jid, first_key, fut
        )
        self.logger.debug("%s maps to %r", key, info)

        return info

    @aioxmpp.service.outbound_presence_filter
    def handle_outbound_presence(self, presence):
        if (presence.type_ == aioxmpp.structs.PresenceType.AVAILABLE
                and self.__active_hashsets):
            current_hashset = self.__active_hashsets[-1]

            try:
                keys = current_hashset[self.__115]
            except KeyError:
                pass
            else:
                self.__115.put_keys(keys, presence)

            try:
                keys = current_hashset[self.__390]
            except KeyError:
                pass
            else:
                self.__390.put_keys(keys, presence)

        return presence

    @aioxmpp.service.inbound_presence_filter
    def handle_inbound_presence(self, presence):
        keys = []

        if self.xep390_support:
            keys.extend(self.__390.extract_keys(presence))

        if self.xep115_support:
            keys.extend(self.__115.extract_keys(presence))

        if keys:
            lookup_task = aioxmpp.utils.LazyTask(
                self.lookup_info,
                presence.from_,
                keys,
            )
            self.disco_client.set_info_future(
                presence.from_,
                None,
                lookup_task
            )

        return presence

    def _push_hashset(self, node, hashset):
        if self.__active_hashsets and hashset == self.__active_hashsets[-1]:
            return False

        for group in hashset.values():
            for key in group:
                if not self.__key_users[key.node]:
                    self.disco_server.mount_node(key.node, node)
                self.__key_users[key.node] += 1
        self.__active_hashsets.append(hashset)

        for expired in self.__active_hashsets[:-3]:
            for group in expired.values():
                for key in group:
                    self.__key_users[key.node] -= 1
                    if not self.__key_users[key.node]:
                        self.disco_server.unmount_node(key.node)
                        del self.__key_users[key.node]

        del self.__active_hashsets[:-3]

        return True

    def update_hash(self):
        node = disco.StaticNode.clone(self.disco_server)
        info = node.as_info_xso()

        new_hashset = {}

        if self.xep115_support:
            new_hashset[self.__115] = set(self.__115.calculate_keys(info))

        if self.xep390_support:
            new_hashset[self.__390] = set(self.__390.calculate_keys(info))

        self.logger.debug("new hashset=%r", new_hashset)

        if self._push_hashset(node, new_hashset):
            self.on_ver_changed()

    # declare those at the bottom so that on_ver_changed gets emitted when the
    # service is instantiated
    _xep115_feature = disco.register_feature(namespaces.xep0115_caps)
    _xep390_feature = disco.register_feature(namespaces.xep0390_caps)


def writeback(path, captured_events):
    aioxmpp.utils.mkdir_exist_ok(path.parent)
    with tempfile.NamedTemporaryFile(dir=str(path.parent),
                                     delete=False) as tmpf:
        try:
            generator = aioxmpp.xml.XMPPXMLGenerator(
                tmpf,
                short_empty_elements=True)
            generator.startDocument()
            aioxmpp.xso.events_to_sax(captured_events, generator)
            generator.endDocument()
        except:  # NOQA
            os.unlink(tmpf.name)
            raise
    os.replace(tmpf.name, str(path))



########################################################################
# File name: errors.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.errors` --- Exception classes
############################################

Exception classes mapping to XMPP stream errors
===============================================

.. autoclass:: StreamError

.. autoclass:: StreamErrorCondition

Exception classes mapping to XMPP stanza errors
===============================================

.. autoclass:: StanzaError

.. autoclass:: XMPPError

.. currentmodule:: aioxmpp

.. autoclass:: ErrorCondition

.. autoclass:: XMPPAuthError

.. autoclass:: XMPPModifyError

.. autoclass:: XMPPCancelError

.. autoclass:: XMPPWaitError

.. autoclass:: XMPPContinueError

.. currentmodule:: aioxmpp.errors

.. autoclass:: ErroneousStanza

Stream negotiation exceptions
=============================

.. autoclass:: StreamNegotiationFailure

.. autoclass:: SecurityNegotiationFailure

.. autoclass:: SASLUnavailable

.. autoclass:: TLSFailure

.. autoclass:: TLSUnavailable

I18N exceptions
===============

.. autoclass:: UserError

.. autoclass:: UserValueError

Other exceptions
================

.. autoclass:: MultiOSError

.. autoclass:: GatherError

"""
import enum
import gettext
import warnings

from . import xso, i18n, structs

from .utils import namespaces


def format_error_text(
        condition,
        text=None,
        application_defined_condition=None):
    error_tag = xso.tag_to_str(condition.value)
    if application_defined_condition is not None:
        error_tag += "/{}".format(
            xso.tag_to_str(application_defined_condition.TAG)
        )
    if text:
        error_tag += " ({!r})".format(text)
    return error_tag


class ErrorCondition(structs.CompatibilityMixin, xso.XSOEnumMixin, enum.Enum):
    """
    Enumeration to represent a :rfc:`6120` stanza error condition. Please
    see :rfc:`6120`, section 8.3.3, for the semantics of the individual
    conditions.

    .. versionadded:: 0.10

    .. attribute:: BAD_REQUEST
        :annotation: = namespaces.stanzas, "bad-request"

    .. attribute:: CONFLICT
        :annotation: = namespaces.stanzas, "conflict"

    .. attribute:: FEATURE_NOT_IMPLEMENTED
        :annotation: = namespaces.stanzas, "feature-not-implemented"

    .. attribute:: FORBIDDEN
        :annotation: = namespaces.stanzas, "forbidden"

    .. attribute:: GONE
        :annotation: = namespaces.stanzas, "gone"

        .. attribute:: xso_class

            .. attribute:: new_address

                The text content of the ``<gone/>`` element represtenting the
                URI at which the entity can now be found.

                May be :data:`None` if there is no such URI.

    .. attribute:: INTERNAL_SERVER_ERROR
        :annotation: = namespaces.stanzas, "internal-server-error"

    .. attribute:: ITEM_NOT_FOUND
        :annotation: = namespaces.stanzas, "item-not-found"

    .. attribute:: JID_MALFORMED
        :annotation: = namespaces.stanzas, "jid-malformed"

    .. attribute:: NOT_ACCEPTABLE
        :annotation: = namespaces.stanzas, "not-acceptable"

    .. attribute:: NOT_ALLOWED
        :annotation: = namespaces.stanzas, "not-allowed"

    .. attribute:: NOT_AUTHORIZED
        :annotation: = namespaces.stanzas, "not-authorized"

    .. attribute:: POLICY_VIOLATION
        :annotation: = namespaces.stanzas, "policy-violation"

    .. attribute:: RECIPIENT_UNAVAILABLE
        :annotation: = namespaces.stanzas, "recipient-unavailable"

    .. attribute:: REDIRECT
        :annotation: = namespaces.stanzas, "redirect"

        .. attribute:: xso_class

            .. attribute:: new_address

                The text content of the ``<redirect/>`` element represtenting
                the URI at which the entity can currently be found.

                May be :data:`None` if there is no such URI.

    .. attribute:: REGISTRATION_REQUIRED
        :annotation: = namespaces.stanzas, "registration-required"

    .. attribute:: REMOTE_SERVER_NOT_FOUND
        :annotation: = namespaces.stanzas, "remote-server-not-found"

    .. attribute:: REMOTE_SERVER_TIMEOUT
        :annotation: = namespaces.stanzas, "remote-server-timeout"

    .. attribute:: RESOURCE_CONSTRAINT
        :annotation: = namespaces.stanzas, "resource-constraint"

    .. attribute:: SERVICE_UNAVAILABLE
        :annotation: = namespaces.stanzas, "service-unavailable"

    .. attribute:: SUBSCRIPTION_REQUIRED
        :annotation: = namespaces.stanzas, "subscription-required"

    .. attribute:: UNDEFINED_CONDITION
        :annotation: = namespaces.stanzas, "undefined-condition"

    .. attribute:: UNEXPECTED_REQUEST
        :annotation: = namespaces.stanzas, "unexpected-request"

    """

    BAD_REQUEST = (namespaces.stanzas, "bad-request")
    CONFLICT = (namespaces.stanzas, "conflict")
    FEATURE_NOT_IMPLEMENTED = (namespaces.stanzas, "feature-not-implemented")
    FORBIDDEN = (namespaces.stanzas, "forbidden")
    GONE = (namespaces.stanzas, "gone")
    INTERNAL_SERVER_ERROR = (namespaces.stanzas, "internal-server-error")
    ITEM_NOT_FOUND = (namespaces.stanzas, "item-not-found")
    JID_MALFORMED = (namespaces.stanzas, "jid-malformed")
    NOT_ACCEPTABLE = (namespaces.stanzas, "not-acceptable")
    NOT_ALLOWED = (namespaces.stanzas, "not-allowed")
    NOT_AUTHORIZED = (namespaces.stanzas, "not-authorized")
    POLICY_VIOLATION = (namespaces.stanzas, "policy-violation")
    RECIPIENT_UNAVAILABLE = (namespaces.stanzas, "recipient-unavailable")
    REDIRECT = (namespaces.stanzas, "redirect")
    REGISTRATION_REQUIRED = (namespaces.stanzas, "registration-required")
    REMOTE_SERVER_NOT_FOUND = (namespaces.stanzas, "remote-server-not-found")
    REMOTE_SERVER_TIMEOUT = (namespaces.stanzas, "remote-server-timeout")
    RESOURCE_CONSTRAINT = (namespaces.stanzas, "resource-constraint")
    SERVICE_UNAVAILABLE = (namespaces.stanzas, "service-unavailable")
    SUBSCRIPTION_REQUIRED = (namespaces.stanzas, "subscription-required")
    UNDEFINED_CONDITION = (namespaces.stanzas, "undefined-condition")
    UNEXPECTED_REQUEST = (namespaces.stanzas, "unexpected-request")


ErrorCondition.GONE.xso_class.new_address = xso.Text()
ErrorCondition.REDIRECT.xso_class.new_address = xso.Text()


class StreamErrorCondition(structs.CompatibilityMixin,
                           xso.XSOEnumMixin,
                           enum.Enum):
    """
    Enumeration to represent a :rfc:`6120` stream  error condition. Please
    see :rfc:`6120`, section 4.9.3, for the semantics of the individual
    conditions.

    .. versionadded:: 0.10

    .. attribute:: BAD_FORMAT
        :annotation: = (namespaces.streams, "bad-format")

    .. attribute:: BAD_NAMESPACE_PREFIX
        :annotation: = (namespaces.streams, "bad-namespace-prefix")

    .. attribute:: CONFLICT
        :annotation: = (namespaces.streams, "conflict")

    .. attribute:: CONNECTION_TIMEOUT
        :annotation: = (namespaces.streams, "connection-timeout")

    .. attribute:: HOST_GONE
        :annotation: = (namespaces.streams, "host-gone")

    .. attribute:: HOST_UNKNOWN
        :annotation: = (namespaces.streams, "host-unknown")

    .. attribute:: IMPROPER_ADDRESSING
        :annotation: = (namespaces.streams, "improper-addressing")

    .. attribute:: INTERNAL_SERVER_ERROR
        :annotation: = (namespaces.streams, "internal-server-error")

    .. attribute:: INVALID_FROM
        :annotation: = (namespaces.streams, "invalid-from")

    .. attribute:: INVALID_NAMESPACE
        :annotation: = (namespaces.streams, "invalid-namespace")

    .. attribute:: INVALID_XML
        :annotation: = (namespaces.streams, "invalid-xml")

    .. attribute:: NOT_AUTHORIZED
        :annotation: = (namespaces.streams, "not-authorized")

    .. attribute:: NOT_WELL_FORMED
        :annotation: = (namespaces.streams, "not-well-formed")

    .. attribute:: POLICY_VIOLATION
        :annotation: = (namespaces.streams, "policy-violation")

    .. attribute:: REMOTE_CONNECTION_FAILED
        :annotation: = (namespaces.streams, "remote-connection-failed")

    .. attribute:: RESET
        :annotation: = (namespaces.streams, "reset")

    .. attribute:: RESOURCE_CONSTRAINT
        :annotation: = (namespaces.streams, "resource-constraint")

    .. attribute:: RESTRICTED_XML
        :annotation: = (namespaces.streams, "restricted-xml")

    .. attribute:: SEE_OTHER_HOST
        :annotation: = (namespaces.streams, "see-other-host")

    .. attribute:: SYSTEM_SHUTDOWN
        :annotation: = (namespaces.streams, "system-shutdown")

    .. attribute:: UNDEFINED_CONDITION
        :annotation: = (namespaces.streams, "undefined-condition")

    .. attribute:: UNSUPPORTED_ENCODING
        :annotation: = (namespaces.streams, "unsupported-encoding")

    .. attribute:: UNSUPPORTED_FEATURE
        :annotation: = (namespaces.streams, "unsupported-feature")

    .. attribute:: UNSUPPORTED_STANZA_TYPE
        :annotation: = (namespaces.streams, "unsupported-stanza-type")

    .. attribute:: UNSUPPORTED_VERSION
        :annotation: = (namespaces.streams, "unsupported-version")

    """

    BAD_FORMAT = (namespaces.streams, "bad-format")
    BAD_NAMESPACE_PREFIX = (namespaces.streams, "bad-namespace-prefix")
    CONFLICT = (namespaces.streams, "conflict")
    CONNECTION_TIMEOUT = (namespaces.streams, "connection-timeout")
    HOST_GONE = (namespaces.streams, "host-gone")
    HOST_UNKNOWN = (namespaces.streams, "host-unknown")
    IMPROPER_ADDRESSING = (namespaces.streams, "improper-addressing")
    INTERNAL_SERVER_ERROR = (namespaces.streams, "internal-server-error")
    INVALID_FROM = (namespaces.streams, "invalid-from")
    INVALID_NAMESPACE = (namespaces.streams, "invalid-namespace")
    INVALID_XML = (namespaces.streams, "invalid-xml")
    NOT_AUTHORIZED = (namespaces.streams, "not-authorized")
    NOT_WELL_FORMED = (namespaces.streams, "not-well-formed")
    POLICY_VIOLATION = (namespaces.streams, "policy-violation")
    REMOTE_CONNECTION_FAILED = (namespaces.streams, "remote-connection-failed")
    RESET = (namespaces.streams, "reset")
    RESOURCE_CONSTRAINT = (namespaces.streams, "resource-constraint")
    RESTRICTED_XML = (namespaces.streams, "restricted-xml")
    SEE_OTHER_HOST = (namespaces.streams, "see-other-host")
    SYSTEM_SHUTDOWN = (namespaces.streams, "system-shutdown")
    UNDEFINED_CONDITION = (namespaces.streams, "undefined-condition")
    UNSUPPORTED_ENCODING = (namespaces.streams, "unsupported-encoding")
    UNSUPPORTED_FEATURE = (namespaces.streams, "unsupported-feature")
    UNSUPPORTED_STANZA_TYPE = (namespaces.streams, "unsupported-stanza-type")
    UNSUPPORTED_VERSION = (namespaces.streams, "unsupported-version")


StreamErrorCondition.SEE_OTHER_HOST.xso_class.new_address = xso.Text()


class StreamError(ConnectionError):
    def __init__(self, condition, text=None):
        if not isinstance(condition, StreamErrorCondition):
            condition = StreamErrorCondition(condition)
            warnings.warn(
                "as of aioxmpp 1.0, stream error conditions must be members of "
                "the aioxmpp.errors.StreamErrorCondition enumeration",
                DeprecationWarning,
                stacklevel=2,
            )

        super().__init__("stream error: {}".format(
            format_error_text(condition, text))
        )
        self.condition = condition
        self.text = text


class StanzaError(Exception):
    pass


class XMPPError(StanzaError):
    """
    Exception representing an error defined in the XMPP protocol.

    :param condition: The :rfc:`6120` defined error condition as enumeration
        member or :class:`aioxmpp.xso.XSO`
    :type condition: :class:`aioxmpp.ErrorCondition` or :class:`aioxmpp.xso.XSO`
    :param text: Optional human-readable text explaining the error
    :type text: :class:`str`
    :param application_defined_condition: Object describing the error in more
        detail
    :type application_defined_condition: :class:`aioxmpp.xso.XSO`

    .. versionchanged:: 0.10

        As of 0.10, `condition` should either be a
        :class:`aioxmpp.ErrorCondition` enumeration member or an XSO
        representing one of the error conditions.

        For compatibility, namespace-localpart tuples indicating the tag of
        the defined error condition are still accepted.

    .. deprecated:: 0.10

        Starting with aioxmpp 1.0, namespace-localpart tuples will not be
        accepted anymore. See the changelog for notes on the transition.

    .. attribute:: condition_obj

        The :class:`aioxmpp.XSO` which represents the error condition.

        .. versionadded:: 0.10

    .. autoattribute:: condition

    .. attribute:: text

        Optional human-readable text describing the error further.

        This is :data:`None` if the text is omitted.

    .. attribute:: application_defined_condition

        Optional :class:`aioxmpp.XSO` which further defines the error
        condition.

    Relevant subclasses:

    .. autosummary::

       aioxmpp.XMPPAuthError
       aioxmpp.XMPPModifyError
       aioxmpp.XMPPCancelError
       aioxmpp.XMPPContinueError
       aioxmpp.XMPPWaitError

    """

    TYPE = structs.ErrorType.CANCEL

    def __init__(self,
                 condition,
                 text=None,
                 application_defined_condition=None):
        if not isinstance(condition, (ErrorCondition, xso.XSO)):
            condition = ErrorCondition(condition)
            warnings.warn(
                "as of aioxmpp 1.0, error conditions must be members of the "
                "aioxmpp.ErrorCondition enumeration",
                DeprecationWarning,
                stacklevel=2,
            )

        super().__init__(format_error_text(
            condition.enum_member,
            text=text,
            application_defined_condition=application_defined_condition))
        self.condition_obj = condition.to_xso()
        self.text = text
        self.application_defined_condition = application_defined_condition

    @property
    def condition(self):
        """
        :class:`aioxmpp.ErrorCondition` enumeration member representing the
        error condition.
        """

        return self.condition_obj.enum_member


class XMPPWarning(XMPPError, UserWarning):
    TYPE = structs.ErrorType.CONTINUE


class XMPPAuthError(XMPPError, PermissionError):
    TYPE = structs.ErrorType.AUTH


class XMPPModifyError(XMPPError, ValueError):
    TYPE = structs.ErrorType.MODIFY


class XMPPCancelError(XMPPError):
    TYPE = structs.ErrorType.CANCEL


class XMPPWaitError(XMPPError):
    TYPE = structs.ErrorType.WAIT


class XMPPContinueError(XMPPWarning):
    TYPE = structs.ErrorType.CONTINUE


class ErroneousStanza(StanzaError):
    """
    This exception is thrown into listeners for IQ responses by
    :class:`aioxmpp.stream.StanzaStream` if a response for an IQ was received,
    but could not be decoded (due to malformed or unsupported payload).

    .. attribute:: partial_obj

       Contains the partially decoded stanza XSO. Do not rely on any members
       except those representing XML attributes (:attr:`~.StanzaBase.to`,
       :attr:`~.StanzaBase.from_`, :attr:`~.StanzaBase.type_`).

    """

    def __init__(self, partial_obj):
        super().__init__("erroneous stanza received: {!r}".format(
            partial_obj))
        self.partial_obj = partial_obj


class StreamNegotiationFailure(ConnectionError):
    pass


class SecurityNegotiationFailure(StreamNegotiationFailure):
    def __init__(self, xmpp_error,
                 kind="Security negotiation failure",
                 text=None):
        msg = "{}: {}".format(kind, xmpp_error)
        if text:
            msg += " ('{}')".format(text)
        super().__init__(msg)
        self.xmpp_error = xmpp_error
        self.text = text


class SASLUnavailable(SecurityNegotiationFailure):
    # we use this to tell the Client that SASL has not been available at all,
    # or that we could not agree on mechansims.
    # it might be helpful to notify the peer about this before dying.
    pass


class TLSFailure(SecurityNegotiationFailure):
    def __init__(self, xmpp_error, text=None):
        super().__init__(xmpp_error, text=text, kind="TLS failure")


class TLSUnavailable(TLSFailure):
    pass


class UserError(Exception):
    """
    An exception subclass, which should be used as a mix-in.

    It is intended to be used for exceptions which may be user-facing, such as
    connection errors, value validation issues and the like.

    `localizable_string` must be a :class:`.i18n.LocalizableString`
    instance. The `args` and `kwargs` will be passed to
    :class:`.LocalizableString.localize` when either :func:`str` is called on
    the :class:`UserError` or :meth:`localize` is called.

    The :func:`str` is created using the default
    :class:`~.i18n.LocalizingFormatter` and a :class:`gettext.NullTranslations`
    instance. The point in time at which the default localizing formatter is
    created is unspecified.

    .. automethod:: localize

    """

    DEFAULT_FORMATTER = i18n.LocalizingFormatter()
    DEFAULT_TRANSLATIONS = gettext.NullTranslations()

    def __init__(self, localizable_string, *args, **kwargs):
        super().__init__()
        self._str = localizable_string.localize(
            self.DEFAULT_FORMATTER,
            self.DEFAULT_TRANSLATIONS,
            *args, **kwargs)
        self.localizable_string = localizable_string
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return str(self._str)

    def localize(self, formatter, translator):
        """
        Return a localized version of the `localizable_string` passed to the
        consturctor. It is formatted using the `formatter` with the `args` and
        `kwargs` passed to the constructor of :class:`UserError`.
        """
        return self.localizable_string.localize(
            formatter,
            translator,
            *self.args,
            **self.kwargs
        )


class UserValueError(UserError, ValueError):
    """
    This is a :class:`ValueError` with :class:`UserError` mixed in.
    """


class MultiOSError(OSError):
    """
    Describe an error situation which has been caused by the sequential
    occurence of multiple other `exceptions`.

    The `message` shall be descriptive and will be prepended to a concatenation
    of the error messages of the given `exceptions`.
    """

    def __init__(self, message, exceptions):
        flattened_exceptions = []
        for exc in exceptions:
            if hasattr(exc, "exceptions"):
                flattened_exceptions.extend(exc.exceptions)
            else:
                flattened_exceptions.append(exc)

        super().__init__(
            "{}: multiple errors: {}".format(
                message,
                ", ".join(map(str, flattened_exceptions))
            )
        )
        self.exceptions = flattened_exceptions


class GatherError(RuntimeError):
    """
    Describe an error situation which has been caused by the occurence
    of multiple other `exceptions`.

    The `message` shall be descriptive and will be prepended to a concatenation
    of the error messages of the given `exceptions`.
    """

    def __init__(self, message, exceptions):
        flattened_exceptions = []
        for exc in exceptions:
            if hasattr(exc, "exceptions"):
                flattened_exceptions.extend(exc.exceptions)
            else:
                flattened_exceptions.append(exc)

        super().__init__(
            "{}: multiple errors: {}".format(
                message,
                ", ".join(map(str, flattened_exceptions))
            )
        )
        self.exceptions = flattened_exceptions



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import contextlib
import unittest

import aioxmpp.dispatcher
import aioxmpp.errors as errors
import aioxmpp.roster.service as roster_service
import aioxmpp.roster.xso as roster_xso
import aioxmpp.service as service
import aioxmpp.stanza as stanza
import aioxmpp.structs as structs

from aioxmpp.utils import namespaces

from aioxmpp.testutils import (
    make_connected_client,
    run_coroutine,
    make_listener,
)


TEST_JID = structs.JID.fromstr("user@foo.example")


class TestItem(unittest.TestCase):
    def setUp(self):
        self.jid = TEST_JID

    def test_init(self):
        item = roster_service.Item(self.jid)
        self.assertEqual(self.jid, item.jid)
        self.assertEqual("none", item.subscription)
        self.assertFalse(item.approved)
        self.assertIsNone(item.ask)
        self.assertIsNone(item.name)
        self.assertSetEqual(set(), item.groups)

        item = roster_service.Item(
            self.jid,
            subscription="both",
            approved=True,
            ask="subscribe",
            name="foobar",
            groups=("fnord", "bar"))
        self.assertEqual("both", item.subscription)
        self.assertTrue(item.approved)
        self.assertEqual("subscribe", item.ask)
        self.assertEqual("foobar", item.name)
        self.assertSetEqual(
            {"fnord", "bar"},
            item.groups
        )

    def test_update_from_xso_item(self):
        xso_item = roster_xso.Item(
            jid=self.jid,
            subscription="to",
            ask="subscribe",
            approved=False,
            name="test",
            groups=[
                roster_xso.Group(name="foo"),
                roster_xso.Group(name="bar"),
            ])

        item = roster_service.Item(self.jid)
        item.update_from_xso_item(xso_item)
        self.assertEqual(xso_item.jid, item.jid)
        self.assertEqual(xso_item.subscription, item.subscription)
        self.assertEqual(xso_item.ask, item.ask)
        self.assertEqual(xso_item.approved, item.approved)
        self.assertEqual(xso_item.name, item.name)
        self.assertSetEqual({"foo", "bar"}, item.groups)

        xso_item = roster_xso.Item(
            jid=structs.JID.fromstr("user@bar.example"),
            subscription="from",
            ask=None,
            approved=True,
            name="other test",
            groups=[
                roster_xso.Group(name="a")
            ])
        item.update_from_xso_item(xso_item)

        self.assertEqual(self.jid, item.jid)
        self.assertEqual(xso_item.subscription, item.subscription)
        self.assertEqual(xso_item.ask, item.ask)
        self.assertEqual(xso_item.approved, item.approved)
        self.assertEqual(xso_item.name, item.name)
        self.assertSetEqual({"a"}, item.groups)

    @unittest.mock.patch.object(roster_service.Item, "update_from_xso_item")
    def test_from_xso_item(self, update_from_xso_item):
        xso_item = roster_xso.Item(
            jid=structs.JID.fromstr("user@bar.example"),
            subscription="from",
            ask=None,
            approved=True)

        item = roster_service.Item.from_xso_item(xso_item)
        self.assertEqual(xso_item.jid, item.jid)
        self.assertSequenceEqual(
            [
                unittest.mock.call(xso_item)
            ],
            update_from_xso_item.mock_calls
        )

    def test_export_as_json(self):
        item = roster_service.Item(
            jid=self.jid,
            subscription="to",
            ask="subscribe",
            approved=False,
            name="test",
            groups=["a", "b"])

        self.assertDictEqual(
            {
                "subscription": "to",
                "ask": "subscribe",
                "name": "test",
                "groups": ["a", "b"],
            },
            item.export_as_json()
        )

        item = roster_service.Item(
            jid=self.jid,
            approved=True,
            groups=["z", "a"])

        self.assertDictEqual(
            {
                "subscription": "none",
                "approved": True,
                "groups": ["a", "z"],
            },
            item.export_as_json()
        )

    def test_update_from_json(self):
        item = roster_service.Item(jid=self.jid)

        item.update_from_json({
            "subscription": "both"
        })
        self.assertEqual("both", item.subscription)

        item.update_from_json({
            "approved": True
        })
        self.assertTrue(item.approved)
        self.assertEqual("none", item.subscription)

        item.update_from_json({
            "ask": "subscribe"
        })
        self.assertEqual("subscribe", item.ask)
        self.assertFalse(item.approved)

        item.update_from_json({
            "name": "foobar baz"
        })
        self.assertEqual("foobar baz", item.name)
        self.assertIsNone(item.ask)

        item.update_from_json({
            "groups": ["a", "b", "a"],
        })
        self.assertIsNone(item.name)
        self.assertSetEqual({"a", "b"}, item.groups)

        item.update_from_json({})
        self.assertSetEqual(set(), item.groups)


class TestService(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.presence_dispatcher = aioxmpp.dispatcher.SimplePresenceDispatcher(
            self.cc
        )
        self.dependencies = {
            aioxmpp.dispatcher.SimplePresenceDispatcher:
                self.presence_dispatcher,
        }
        self.s = roster_service.RosterClient(
            self.cc,
            dependencies=self.dependencies
        )

        self.user1 = structs.JID.fromstr("user@foo.example")
        self.user2 = structs.JID.fromstr("user@bar.example")

        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user1,
                    groups=[
                        roster_xso.Group(name="group1"),
                        roster_xso.Group(name="group3"),
                    ]
                ),
                roster_xso.Item(
                    jid=self.user2,
                    name="some bar user",
                    subscription="both",
                    groups=[
                        roster_xso.Group(name="group1"),
                        roster_xso.Group(name="group2"),
                    ]
                )
            ],
            ver="foobar"
        )

        self.cc.send.return_value = response

        run_coroutine(self.cc.before_stream_established())

        self.cc.send.reset_mock()

        self.listener = make_listener(self.s)

    def test_is_Service(self):
        self.assertIsInstance(
            self.s,
            service.Service
        )

    def test_init(self):
        # required to clear listeners of dependency
        run_coroutine(self.s.shutdown())
        s = roster_service.RosterClient(
            self.cc,
            dependencies=self.dependencies
        )
        self.assertDictEqual({}, s.items)
        self.assertEqual(None, s.version)
        self.assertDictEqual({}, s.groups)

    def test_handle_roster_push_is_decorated(self):
        self.assertTrue(
            service.is_iq_handler(
                structs.IQType.SET,
                roster_xso.Query,
                roster_service.RosterClient.handle_roster_push,
            )
        )

    def test_handle_subscribe_is_decorated(self):
        self.assertTrue(
            aioxmpp.dispatcher.is_presence_handler(
                structs.PresenceType.SUBSCRIBE,
                None,
                roster_service.RosterClient.handle_subscribe,
            )
        )

    def test_handle_subscribed_is_decorated(self):
        self.assertTrue(
            aioxmpp.dispatcher.is_presence_handler(
                structs.PresenceType.SUBSCRIBED,
                None,
                roster_service.RosterClient.handle_subscribed,
            )
        )

    def test_handle_unsubscribe_is_decorated(self):
        self.assertTrue(
            aioxmpp.dispatcher.is_presence_handler(
                structs.PresenceType.UNSUBSCRIBE,
                None,
                roster_service.RosterClient.handle_unsubscribe,
            )
        )

    def test_handle_unsubscribed_is_decorated(self):
        self.assertTrue(
            aioxmpp.dispatcher.is_presence_handler(
                structs.PresenceType.UNSUBSCRIBED,
                None,
                roster_service.RosterClient.handle_unsubscribed,
            )
        )

    def test_request_initial_roster_before_stream_established(self):
        self.assertIn(self.user1, self.s.items)
        self.assertIn(self.user2, self.s.items)
        self.assertEqual("foobar", self.s.version)

        self.assertEqual("both", self.s.items[self.user2].subscription)
        self.assertEqual("some bar user", self.s.items[self.user2].name)

    def test_handle_roster_push_rejects_push_with_nonempty_from(self):
        self.cc.local_jid = structs.JID.fromstr("foo@bar.example")

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.from_ = structs.JID.fromstr("fnord@bar.example")

        with self.assertRaises(errors.XMPPAuthError) as ctx:
            run_coroutine(self.s.handle_roster_push(iq))

        self.assertEqual(
            errors.ErrorCondition.FORBIDDEN,
            ctx.exception.condition
        )

    def test_handle_roster_push_accepts_push_from_bare_local_jid(self):
        self.cc.local_jid = structs.JID.fromstr("foo@bar.example/fnord")

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.from_ = structs.JID.fromstr("foo@bar.example")
        iq.payload = roster_xso.Query()

        run_coroutine(self.s.handle_roster_push(iq))

    def test_handle_roster_push_rejects_push_from_full_local_jid(self):
        self.cc.local_jid = structs.JID.fromstr("foo@bar.example/fnord")

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.from_ = structs.JID.fromstr("foo@bar.example/fnord")
        iq.payload = roster_xso.Query()

        with self.assertRaises(errors.XMPPAuthError) as ctx:
            run_coroutine(self.s.handle_roster_push(iq))

        self.assertEqual(
            errors.ErrorCondition.FORBIDDEN,
            ctx.exception.condition
        )

    def test_handle_roster_push_extends_roster(self):
        user1 = structs.JID.fromstr("user2@foo.example")
        user2 = structs.JID.fromstr("user2@bar.example")

        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=user1),
                roster_xso.Item(
                    jid=user2,
                    name="some bar user",
                    subscription="both"
                )
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        self.assertIsNone(
            run_coroutine(self.s.handle_roster_push(iq))
        )

        self.assertIn(user1, self.s.items)
        self.assertIn(user2, self.s.items)
        self.assertEqual("foobar", self.s.version)

        self.assertEqual("both", self.s.items[user2].subscription)
        self.assertEqual("some bar user", self.s.items[user2].name)

    def test_handle_roster_push_removes_from_roster(self):
        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user1,
                    subscription="remove"),
            ],
            ver="foobarbaz"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        self.assertIsNone(
            run_coroutine(self.s.handle_roster_push(iq))
        )

        self.assertNotIn(self.user1, self.s.items)
        self.assertIn(self.user2, self.s.items)
        self.assertEqual("foobarbaz", self.s.version)

    def test_item_objects_do_not_change_during_push(self):
        old_item = self.s.items[self.user1]

        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user1,
                    subscription="both"
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        self.assertIsNone(
            run_coroutine(self.s.handle_roster_push(iq))
        )

        self.assertIs(old_item, self.s.items[self.user1])
        self.assertEqual("both", old_item.subscription)

    def test_initial_roster_discards_information(self):
        self.cc.mock_calls.clear()

        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="some bar user",
                    subscription="both"
                )
            ],
            ver="foobar"
        )

        self.cc.send.return_value = response

        run_coroutine(self.cc.before_stream_established())
        self.assertSequenceEqual(
            [
                unittest.mock.call.send(
                    unittest.mock.ANY,
                    timeout=self.cc.negotiation_timeout.total_seconds()
                )
            ],
            self.cc.mock_calls
        )

        self.assertNotIn(self.user1, self.s.items)

    def test_initial_roster_fires_event(self):
        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="some bar user",
                    subscription="both"
                )
            ],
            ver="foobar"
        )

        cb = unittest.mock.Mock()
        cb.return_value = True

        def cb_impl():
            cb()
            # assure that the roster update is already finished
            self.assertNotIn(self.user1, self.s.items)

        self.s.on_initial_roster_received.connect(cb_impl)

        self.cc.send.return_value = response
        self.cc.send.delay = 0.05

        task = asyncio.ensure_future(self.cc.before_stream_established())

        run_coroutine(asyncio.sleep(0.01))

        self.assertSequenceEqual([], cb.mock_calls)

        run_coroutine(asyncio.sleep(0.041))

        self.assertSequenceEqual(
            [
                unittest.mock.call(),
            ],
            cb.mock_calls
        )

    def test_initial_roster_fires_group_event(self):
        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="some bar user",
                    subscription="both",
                    groups={roster_xso.Group(name="a"),
                            roster_xso.Group(name="b")}
                ),
                roster_xso.Item(
                    jid=self.user1,
                    name="some foo user",
                    subscription="both",
                    groups={roster_xso.Group(name="a"),
                            roster_xso.Group(name="c")}
                )
            ],
            ver="foobar"
        )

        self.cc.send.return_value = response

        run_coroutine(self.cc.before_stream_established())

        self.assertCountEqual(
            self.listener.on_group_added.mock_calls,
            [
                unittest.mock.call("a"),
                unittest.mock.call("b"),
                unittest.mock.call("c"),
            ]
        )

    def test_initial_roster_does_not_emit_entry_added_for_existing(self):
        old_item = self.s.items[self.user2]

        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="new name",
                    subscription="both"
                ),
                roster_xso.Item(
                    jid=self.user2.replace(localpart="user2"),
                    name="other name",
                )
            ],
            ver="foobar"
        )

        mock = unittest.mock.Mock()
        mock.return_value = False
        self.s.on_entry_added.connect(mock)

        self.cc.send.return_value = response

        run_coroutine(self.cc.before_stream_established())

        self.assertSequenceEqual(
            [
                unittest.mock.call(
                    self.s.items[self.user2.replace(localpart="user2")]
                ),
            ],
            mock.mock_calls
        )

    def test_initial_roster_keeps_existing_entries_alive(self):
        old_item = self.s.items[self.user2]

        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="new name",
                    subscription="both"
                )
            ],
            ver="foobar"
        )

        self.cc.send.return_value = response

        run_coroutine(self.cc.before_stream_established())

        self.assertIs(old_item, self.s.items[self.user2])
        self.assertEqual("new name", old_item.name)

    def test_initial_roster_removes_contact_from_groups(self):
        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="some bar user",
                    subscription="both",
                    groups=[
                        roster_xso.Group(name="group1"),
                        roster_xso.Group(name="group2"),
                    ]
                )
            ],
            ver="foobar"
        )

        self.cc.send.return_value = response

        run_coroutine(self.cc.before_stream_established())

        self.assertSetEqual(
            self.s.groups["group1"],
            {self.s.items[self.user2]},
        )

        self.assertSetEqual(
            self.s.groups["group2"],
            {self.s.items[self.user2]},
        )

        self.assertSetEqual(
            self.s.groups.get("group3", set()),
            set(),
        )

    def test_initial_roster_fires_group_removed_event_for_removed_contact(self):  # NOQA
        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="some bar user",
                    subscription="both",
                    groups=[
                        roster_xso.Group(name="group1"),
                        roster_xso.Group(name="group2"),
                    ]
                )
            ],
            ver="foobar"
        )

        self.cc.send.return_value = response

        run_coroutine(self.cc.before_stream_established())

        self.assertCountEqual(
            self.listener.on_group_removed.mock_calls,
            [
                unittest.mock.call("group3"),
            ]
        )

    def test_initial_roster_fires_group_removed_event_for_changed_contact(self):  # NOQA
        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="some bar user",
                    subscription="both",
                    groups=[
                        roster_xso.Group(name="group1"),
                        roster_xso.Group(name="group2"),
                    ]
                ),
                roster_xso.Item(
                    jid=self.user1,
                    name="some foo user",
                    subscription="both",
                    groups={roster_xso.Group(name="group1")}
                )
            ],
            ver="foobar"
        )

        self.cc.send.return_value = response

        run_coroutine(self.cc.before_stream_established())

        self.assertCountEqual(
            self.listener.on_group_removed.mock_calls,
            [
                unittest.mock.call("group3"),
            ]
        )

    def test_groups_are_cleaned_up_up_when_on_entry_removed_fires_on_initial_roster(self):  # NOQA
        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="some bar user",
                    subscription="both",
                    groups=[
                        roster_xso.Group(name="group1"),
                        roster_xso.Group(name="group2"),
                    ]
                ),
            ],
            ver="foobar"
        )

        fut = asyncio.Future()

        def handler(item):
            try:
                for group in item.groups:
                    try:
                        members = self.s.groups[group]
                    except KeyError:
                        members = set()
                    self.assertNotIn(item, members)

            except Exception as exc:
                fut.set_exception(exc)
            else:
                fut.set_result(None)

        self.cc.send.return_value = response

        self.s.on_entry_removed.connect(handler)

        run_coroutine(self.cc.before_stream_established())

        run_coroutine(fut)

    def test_on_entry_name_changed(self):
        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user1,
                    name="foobarbaz",
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        cb = unittest.mock.Mock()
        with self.s.on_entry_name_changed.context_connect(cb):
            run_coroutine(self.s.handle_roster_push(iq))
            run_coroutine(self.s.handle_roster_push(iq))

        self.assertSequenceEqual(
            [
                unittest.mock.call(self.s.items[self.user1]),
            ],
            cb.mock_calls
        )

    def test_on_entry_subscription_state_changed(self):
        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user1,
                    subscription="both",
                    approved=True,
                    ask="subscribe"
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        cb = unittest.mock.Mock()
        with self.s.on_entry_subscription_state_changed.context_connect(cb):
            run_coroutine(self.s.handle_roster_push(iq))
            run_coroutine(self.s.handle_roster_push(iq))

        self.assertSequenceEqual(
            [
                unittest.mock.call(self.s.items[self.user1]),
            ],
            cb.mock_calls
        )

    def test_on_entry_removed(self):
        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user1,
                    subscription="remove",
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        old_item = self.s.items[self.user1]

        cb = unittest.mock.Mock()
        with self.s.on_entry_removed.context_connect(cb):
            run_coroutine(self.s.handle_roster_push(iq))
            run_coroutine(self.s.handle_roster_push(iq))

        self.assertSequenceEqual(
            [
                unittest.mock.call(old_item),
            ],
            cb.mock_calls
        )

    def test_on_entry_added(self):
        new_jid = structs.JID.fromstr("fnord@foo.example")

        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=new_jid,
                    subscription="none",
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        cb = unittest.mock.Mock()
        with self.s.on_entry_added.context_connect(cb):
            run_coroutine(self.s.handle_roster_push(iq))
            run_coroutine(self.s.handle_roster_push(iq))

        self.assertSequenceEqual(
            [
                unittest.mock.call(self.s.items[new_jid]),
            ],
            cb.mock_calls
        )

    def test_on_entry_removed_called_from_initial_roster(self):
        response = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="some bar user",
                    subscription="both"
                )
            ],
            ver="foobar"
        )

        old_item = self.s.items[self.user1]

        self.cc.send.return_value = response

        cb = unittest.mock.Mock()
        with self.s.on_entry_removed.context_connect(cb):
            run_coroutine(self.cc.before_stream_established())

        self.assertSequenceEqual(
            [
                unittest.mock.call(old_item),
            ],
            cb.mock_calls
        )

    def test_on_group_added_for_new_contact(self):
        new_jid = structs.JID.fromstr("fnord@foo.example")

        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=new_jid,
                    subscription="none",
                    groups={
                        roster_xso.Group(name="a"),
                        roster_xso.Group(name="group1"),
                    },
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        run_coroutine(self.s.handle_roster_push(iq))

        self.listener.on_group_added.assert_called_once_with("a")

    def test_groups_are_set_up_when_on_entry_added_fires(self):
        fut = asyncio.Future()

        def handler(item):
            try:
                for group in item.groups:
                    self.assertIn(group, self.s.groups)
                    self.assertIn(item, self.s.groups[group])

            except Exception as exc:
                fut.set_exception(exc)
            else:
                fut.set_result(None)

        new_jid = structs.JID.fromstr("fnord@foo.example")

        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=new_jid,
                    subscription="none",
                    groups={
                        roster_xso.Group(name="a"),
                        roster_xso.Group(name="group1"),
                    },
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        self.s.on_entry_added.connect(handler)

        run_coroutine(self.s.handle_roster_push(iq))

        run_coroutine(fut)

    def test_on_group_added_for_existing_contact(self):
        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    subscription="none",
                    groups={
                        roster_xso.Group(name="group1"),
                        roster_xso.Group(name="group2"),
                        roster_xso.Group(name="group4"),
                    },
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        run_coroutine(self.s.handle_roster_push(iq))

        self.listener.on_group_added.assert_called_once_with("group4")

    def test_on_group_removed_for_existing_contact(self):
        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    subscription="none",
                    groups={
                        roster_xso.Group(name="group1"),
                    },
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        run_coroutine(self.s.handle_roster_push(iq))

        self.listener.on_group_removed.assert_called_once_with("group2")

    def test_on_group_removed_for_removed_contact(self):
        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    subscription="remove",
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        run_coroutine(self.s.handle_roster_push(iq))

        self.listener.on_group_removed.assert_called_once_with("group2")

    def test_groups_are_cleaned_up_up_when_on_entry_removed_fires_on_push(self):
        fut = asyncio.Future()

        def handler(item):
            try:
                for group in item.groups:
                    try:
                        members = self.s.groups[group]
                    except KeyError:
                        members = set()
                    self.assertNotIn(item, members)

            except Exception as exc:
                fut.set_exception(exc)
            else:
                fut.set_result(None)

        new_jid = structs.JID.fromstr("fnord@foo.example")

        request = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    subscription="remove",
                ),
            ],
            ver="foobar"
        )

        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = request

        self.s.on_entry_removed.connect(handler)

        run_coroutine(self.s.handle_roster_push(iq))

        run_coroutine(fut)

    def test_export_as_json(self):
        self.assertDictEqual(
            {
                "items": {
                    str(self.user1): {
                        "subscription": "none",
                        "groups": ["group1", "group3"],
                    },
                    str(self.user2): {
                        "subscription": "both",
                        "name": "some bar user",
                        "groups": ["group1", "group2"],
                    },
                },
                "ver": "foobar",
            },
            self.s.export_as_json()
        )

    def test_import_from_json(self):
        jid1 = structs.JID.fromstr("fnord@foo.example")
        jid2 = structs.JID.fromstr("fnord@bar.example")

        data = {
            "items": {
                str(jid1): {
                    "name": "foo fnord",
                    "subscription": "both",
                },
                str(jid2): {
                    "name": "bar fnord",
                    "subscription": "to",
                }
            },
            "ver": "foobarbaz",
        }

        # import_from_json does not fire events

        cb = unittest.mock.Mock()
        with contextlib.ExitStack() as stack:
            stack.enter_context(
                self.s.on_entry_added.context_connect(cb)
            )
            stack.enter_context(
                self.s.on_entry_removed.context_connect(cb)
            )
            self.s.import_from_json(data)

        self.assertEqual("foobarbaz", self.s.version)

        self.assertNotIn(self.user1, self.s.items)
        self.assertNotIn(self.user2, self.s.items)

        self.assertIn(jid1, self.s.items)
        self.assertIn(jid2, self.s.items)

        self.assertEqual(self.s.items[jid1].name, "foo fnord")
        self.assertEqual(self.s.items[jid1].subscription, "both")

        self.assertEqual(self.s.items[jid2].name, "bar fnord")
        self.assertEqual(self.s.items[jid2].subscription, "to")

        self.assertSequenceEqual([], cb.mock_calls)

    def test_do_not_send_versioned_request_if_not_supported_by_server(self):
        response = roster_xso.Query()

        self.cc.send.return_value = response

        run_coroutine(self.cc.before_stream_established())

        call, = self.cc.send.mock_calls
        _, call_args, call_kwargs = call

        iq_request, = call_args
        self.assertIsNone(
            iq_request.payload.ver
        )

        self.assertNotIn(self.user1, self.s.items)
        self.assertNotIn(self.user2, self.s.items)

    def test_send_versioned_request_if_not_supported_by_server(self):
        self.cc.stream_features[...] = roster_xso.RosterVersioningFeature()

        response = roster_xso.Query()

        self.cc.send.return_value = response

        run_coroutine(self.cc.before_stream_established())

        call, = self.cc.send.mock_calls
        _, call_args, call_kwargs = call

        iq_request, = call_args
        self.assertEqual(
            "foobar",
            iq_request.payload.ver
        )

    def test_process_none_response_to_versioned_request(self):
        self.cc.stream_features[...] = roster_xso.RosterVersioningFeature()

        self.cc.send.return_value = None

        cb = unittest.mock.Mock()
        cb.return_value = True

        self.s.on_initial_roster_received.connect(cb)

        run_coroutine(self.cc.before_stream_established())

        call, = self.cc.send.mock_calls
        _, call_args, call_kwargs = call

        iq_request, = call_args
        self.assertEqual(
            "foobar",
            iq_request.payload.ver
        )

        self.assertIn(self.user1, self.s.items)
        self.assertIn(self.user2, self.s.items)

        self.assertSequenceEqual(
            [
                unittest.mock.call(),
            ],
            cb.mock_calls
        )

    def test_groups_in_initial_roster(self):
        self.assertIn("group1", self.s.groups)
        self.assertIn("group2", self.s.groups)
        self.assertIn("group3", self.s.groups)

        self.assertSetEqual(
            {
                self.s.items[self.user1],
                self.s.items[self.user2],
            },
            self.s.groups["group1"]
        )

        self.assertSetEqual(
            {
                self.s.items[self.user1],
            },
            self.s.groups["group3"]
        )

        self.assertSetEqual(
            {
                self.s.items[self.user2],
            },
            self.s.groups["group2"]
        )

    def test_update_groups_on_update(self):
        request = roster_xso.Query(items=[
            roster_xso.Item(
                jid=self.user1,
                groups=[
                    roster_xso.Group(name="group4")
                ],
            )
        ])

        run_coroutine(self.s.handle_roster_push(
            stanza.IQ(
                structs.IQType.SET,
                payload=request
            )
        ))

        self.assertNotIn("group3", self.s.groups)
        self.assertSetEqual(
            {self.s.items[self.user2]},
            self.s.groups["group1"]
        )
        self.assertSetEqual(
            {self.s.items[self.user2]},
            self.s.groups["group2"]
        )
        self.assertSetEqual(
            {self.s.items[self.user1]},
            self.s.groups["group4"]
        )

    def test_groups_update_fires_events(self):
        request = roster_xso.Query(items=[
            roster_xso.Item(
                jid=self.user1,
                groups=[
                    roster_xso.Group(name="group4")
                ],
            )
        ])

        added_cb = unittest.mock.Mock()
        added_cb.return_value = False
        removed_cb = unittest.mock.Mock()
        removed_cb.return_value = False

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                self.s.on_entry_added_to_group.context_connect(added_cb)
            )
            stack.enter_context(
                self.s.on_entry_removed_from_group.context_connect(removed_cb)
            )
            run_coroutine(self.s.handle_roster_push(
                stanza.IQ(
                    structs.IQType.SET,
                    payload=request
                )
            ))

        self.assertSequenceEqual(
            [
                unittest.mock.call(self.s.items[self.user1], "group4"),
            ],
            added_cb.mock_calls
        )

        self.assertIn(
            unittest.mock.call(self.s.items[self.user1], "group1"),
            removed_cb.mock_calls
        )

        self.assertIn(
            unittest.mock.call(self.s.items[self.user1], "group3"),
            removed_cb.mock_calls
        )

    def test_import_from_json_fixes_groups(self):
        self.s.import_from_json({
            "items": {
                str(self.user1): {
                    "groups": ["a", "b"],
                },
                str(self.user2): {
                    "groups": ["b", "c"],
                }
            }
        })

        self.assertSetEqual(
            set("abc"),
            set(self.s.groups.keys())
        )

        self.assertSetEqual(
            {self.s.items[self.user1], self.s.items[self.user2]},
            self.s.groups["b"]
        )

        self.assertSetEqual(
            {self.s.items[self.user1]},
            self.s.groups["a"]
        )

        self.assertSetEqual(
            {self.s.items[self.user2]},
            self.s.groups["c"]
        )

    def test_item_removal_fixes_groups(self):
        request = roster_xso.Query(items=[
            roster_xso.Item(
                jid=self.user1,
                subscription="remove",
            )
        ])

        added_cb = unittest.mock.Mock()
        added_cb.return_value = False
        removed_cb = unittest.mock.Mock()
        removed_cb.return_value = False

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                self.s.on_entry_added_to_group.context_connect(added_cb)
            )
            stack.enter_context(
                self.s.on_entry_removed_from_group.context_connect(removed_cb)
            )
            run_coroutine(self.s.handle_roster_push(
                stanza.IQ(
                    structs.IQType.SET,
                    payload=request
                )
            ))

        self.assertSequenceEqual([], added_cb.mock_calls)
        self.assertSequenceEqual([], removed_cb.mock_calls)

        self.assertSetEqual(
            {"group1", "group2"},
            set(self.s.groups.keys())
        )

        self.assertSetEqual(
            {self.s.items[self.user2]},
            self.s.groups["group1"]
        )

        self.assertSetEqual(
            {self.s.items[self.user2]},
            self.s.groups["group2"]
        )

    def test_set_entry_name(self):
        self.cc.send.return_value = None

        run_coroutine(
            self.s.set_entry(
                self.user1,
                name="foobar",
                timeout=10
            ),
        )

        self.assertSequenceEqual(
            [
                unittest.mock.call(unittest.mock.ANY, timeout=10)
            ],
            self.cc.send.mock_calls
        )

        call, = self.cc.send.mock_calls
        _, call_args, _ = call

        request_iq, = call_args
        self.assertIsNone(request_iq.to)
        self.assertIsInstance(
            request_iq.payload,
            roster_xso.Query
        )

        query = request_iq.payload
        self.assertIsNone(query.ver)
        self.assertEqual(1, len(query.items))

        item, = query.items
        self.assertEqual(
            self.user1,
            item.jid
        )
        self.assertEqual(
            "foobar",
            item.name
        )
        self.assertSetEqual(
            {
                "group1", "group3"
            },
            {group.name for group in item.groups}
        )

        # defaults
        self.assertEqual("none", item.subscription)
        self.assertFalse(item.approved)
        self.assertIsNone(item.ask)

    def test_set_entry_groups(self):
        self.cc.send.return_value = None

        run_coroutine(
            self.s.set_entry(
                self.user2,
                add_to_groups={"a", "b"},
                remove_from_groups={"group1"},
                timeout=10
            )
        )

        self.assertSequenceEqual(
            [
                unittest.mock.call(unittest.mock.ANY, timeout=10)
            ],
            self.cc.send.mock_calls
        )

        call, = self.cc.send.mock_calls
        _, call_args, _ = call

        request_iq, = call_args
        self.assertIsNone(request_iq.to)
        self.assertIsInstance(
            request_iq.payload,
            roster_xso.Query
        )

        query = request_iq.payload
        self.assertIsNone(query.ver)
        self.assertEqual(1, len(query.items))

        item, = query.items
        self.assertEqual(
            self.user2,
            item.jid
        )
        self.assertEqual(
            "some bar user",
            item.name
        )
        self.assertSetEqual(
            {
                "a", "b", "group2"
            },
            {group.name for group in item.groups}
        )

        # defaults
        self.assertEqual("none", item.subscription)
        self.assertFalse(item.approved)
        self.assertIsNone(item.ask)

    def test_remove_entry(self):
        self.cc.send.return_value = None

        run_coroutine(
            self.s.remove_entry(
                self.user2,
                timeout=10
            )
        )

        self.assertSequenceEqual(
            [
                unittest.mock.call(unittest.mock.ANY, timeout=10)
            ],
            self.cc.send.mock_calls
        )

        call, = self.cc.send.mock_calls
        _, call_args, _ = call

        request_iq, = call_args
        self.assertIsNone(request_iq.to)
        self.assertIsInstance(
            request_iq.payload,
            roster_xso.Query
        )

        query = request_iq.payload
        self.assertIsNone(query.ver)
        self.assertEqual(1, len(query.items))

        item, = query.items
        self.assertEqual(
            self.user2,
            item.jid
        )
        self.assertEqual("remove", item.subscription)

        # defaults
        self.assertIsNone(item.ask)
        self.assertFalse(item.approved)
        self.assertFalse(item.groups)
        self.assertIsNone(item.name)

    def test_handle_subscribe_emits_event(self):
        st = stanza.Presence(
            type_=structs.PresenceType.SUBSCRIBE,
            from_=TEST_JID
        )

        mock = unittest.mock.Mock()
        self.s.on_subscribe.connect(mock)
        self.s.handle_subscribe(st)
        self.assertSequenceEqual(
            mock.mock_calls,
            [
                unittest.mock.call(st)
            ]
        )

    def test_handle_subscribed_emits_event(self):
        st = stanza.Presence(
            type_=structs.PresenceType.SUBSCRIBED,
            from_=TEST_JID
        )

        mock = unittest.mock.Mock()
        self.s.on_subscribed.connect(mock)
        self.s.handle_subscribed(st)
        self.assertSequenceEqual(
            mock.mock_calls,
            [
                unittest.mock.call(st)
            ]
        )

    def test_handle_unsubscribed_emits_event(self):
        st = stanza.Presence(
            type_=structs.PresenceType.UNSUBSCRIBED,
            from_=TEST_JID
        )

        mock = unittest.mock.Mock()
        self.s.on_unsubscribed.connect(mock)
        self.s.handle_unsubscribed(st)
        self.assertSequenceEqual(
            mock.mock_calls,
            [
                unittest.mock.call(st)
            ]
        )

    def test_handle_unsubscribe_emits_event(self):
        st = stanza.Presence(
            type_=structs.PresenceType.UNSUBSCRIBE,
            from_=TEST_JID
        )

        mock = unittest.mock.Mock()
        self.s.on_unsubscribe.connect(mock)
        self.s.handle_unsubscribe(st)
        self.assertSequenceEqual(
            mock.mock_calls,
            [
                unittest.mock.call(st)
            ]
        )

    def test_approve_sends_subscribed_presence(self):
        self.s.approve(TEST_JID)

        self.assertSequenceEqual(
            self.cc.enqueue.mock_calls,
            [
                unittest.mock.call(unittest.mock.ANY),
            ]
        )

        call, = self.cc.enqueue.mock_calls
        _, call_args, _ = call

        st, = call_args
        self.assertIsInstance(st, stanza.Presence)
        self.assertEqual(st.to, TEST_JID)
        self.assertEqual(st.type_, structs.PresenceType.SUBSCRIBED)

    def test_subscribe_sends_subscribe_presence(self):
        self.s.subscribe(TEST_JID)

        self.assertSequenceEqual(
            self.cc.enqueue.mock_calls,
            [
                unittest.mock.call(unittest.mock.ANY),
            ]
        )

        call, = self.cc.enqueue.mock_calls
        _, call_args, _ = call

        st, = call_args
        self.assertIsInstance(st, stanza.Presence)
        self.assertEqual(st.to, TEST_JID)
        self.assertEqual(st.type_, structs.PresenceType.SUBSCRIBE)

    def test_unsubscribe_sends_unsubscribe_presence(self):
        self.s.unsubscribe(TEST_JID)

        self.assertSequenceEqual(
            self.cc.enqueue.mock_calls,
            [
                unittest.mock.call(unittest.mock.ANY),
            ]
        )

        call, = self.cc.enqueue.mock_calls
        _, call_args, _ = call

        st, = call_args
        self.assertIsInstance(st, stanza.Presence)
        self.assertEqual(st.to, TEST_JID)
        self.assertEqual(st.type_, structs.PresenceType.UNSUBSCRIBE)

    def test_do_not_lose_update_during_initial_roster(self):
        self.cc.mock_calls.clear()

        initial = roster_xso.Query(
            items=[
                roster_xso.Item(
                    jid=self.user2,
                    name="some bar user",
                    subscription="both"
                )
            ],
            ver="foobar"
        )

        push = stanza.IQ(
            type_=structs.IQType.SET,
            payload=roster_xso.Query(
                items=[
                    roster_xso.Item(
                        jid=self.user1,
                        name="some foo user",
                    ),
                    roster_xso.Item(
                        jid=self.user2,
                        subscription="remove",
                    )
                ],
                ver="foobar"
            )
        )

        @asyncio.coroutine
        def send(iq, timeout=None):
            # this is brutal, but a sure way to provoke the race
            asyncio.ensure_future(self.s.handle_roster_push(push))
            # give the roster push a chance to act
            # (we cannot yield from the handle_roster_push() here: in the fixed
            # version that would be a deadlock)
            yield from asyncio.sleep(0)
            return initial

        self.cc.send = unittest.mock.Mock()
        self.cc.send.side_effect = send

        initial_roster = asyncio.ensure_future(
            self.cc.before_stream_established()
        )

        run_coroutine(initial_roster)

        self.assertNotIn(
            self.user2,
            self.s.items,
            "initial roster processing lost a race against roster push"
        )
